import re
from copy import deepcopy
from unittest.mock import MagicMock

import pandas as pd

from blueetl import features as test_module
from blueetl.config.analysis_model import FeaturesConfig
from blueetl.extract.feature import Feature
from blueetl.utils import ensure_dtypes
from tests.unit.utils import assert_frame_equal


def test_concatenated_features():
    base_df = pd.DataFrame(
        {
            "simulation_id": [0, 0, 0, 0],
            "circuit_id": [0, 0, 0, 0],
            "neuron_class": ["EXC", "EXC", "INH", "INH"],
            "window": ["w0", "w1", "w0", "w1"],
            "decay": [12.85, 9.85, 8.60, 13.40],
            "ratio": [0.10, 0.15, 0.50, 0.55],
        }
    ).set_index(["simulation_id", "circuit_id", "neuron_class", "window"])

    parent = MagicMock()
    obj = test_module.ConcatenatedFeatures(parent)
    assert {"params", "aliases", "df"}.isdisjoint(obj.__dict__)

    base_config = FeaturesConfig(
        type="multi",
        groupby=["simulation_id", "circuit_id", "neuron_class", "window"],
        function="mymodule.myfunc",
        neuron_classes=["EXC", "INH"],
        windows=["w0", "w1"],
    )
    size = 2
    for i in range(size):
        df = base_df.copy(deep=True)
        df["decay"] = df["decay"] * (i + 1)
        feature = Feature.from_pandas(df)
        config = deepcopy(base_config)
        config.params = {"common": 123, "nested": {"myparam1": i * 10, "myparam2": [0.0, 0.5]}}
        config.suffix = f"_{i}"
        name = f"myfeature{config.suffix}"
        setattr(parent, name, feature)
        obj.update(feature_name=name, features_config=config)

    expected_params = pd.DataFrame(
        {
            "common": [123, 123],
            "nested.myparam1": [0, 10],
            "nested.myparam2": [[0.0, 0.5], [0.0, 0.5]],
        },
        index=pd.RangeIndex(size, name="params_id"),
    )
    assert_frame_equal(obj.params, expected_params)

    expected_aliases = pd.DataFrame(
        {
            "column": ["nested.myparam1"],
            "alias": ["myparam1"],
        }
    )
    assert_frame_equal(obj.aliases, expected_aliases)

    expected_df = pd.DataFrame(
        {
            "simulation_id": [0, 0, 0, 0, 0, 0, 0, 0],
            "circuit_id": [0, 0, 0, 0, 0, 0, 0, 0],
            "neuron_class": ["EXC", "EXC", "INH", "INH", "EXC", "EXC", "INH", "INH"],
            "window": ["w0", "w1", "w0", "w1", "w0", "w1", "w0", "w1"],
            "decay": [12.85, 9.85, 8.60, 13.40, 25.7, 19.7, 17.2, 26.8],
            "ratio": [0.10, 0.15, 0.50, 0.55, 0.10, 0.15, 0.50, 0.55],
            "params_id": [0, 0, 0, 0, 1, 1, 1, 1],
            "myparam1": [0, 0, 0, 0, 10, 10, 10, 10],
        }
    ).set_index(["simulation_id", "circuit_id", "neuron_class", "window"])
    expected_df = ensure_dtypes(expected_df)
    assert_frame_equal(obj.df, expected_df)

    # test clear_cache
    assert {"params", "aliases", "df"}.issubset(obj.__dict__)
    obj.clear_cache()
    assert {"params", "aliases", "df"}.isdisjoint(obj.__dict__)
    assert_frame_equal(obj.params, expected_params)
    assert_frame_equal(obj.aliases, expected_aliases)
    assert_frame_equal(obj.df, expected_df)
    assert {"params", "aliases", "df"}.issubset(obj.__dict__)

    # test cloning
    parent = MagicMock()
    clone = obj.clone(parent=parent)
    assert clone._parent == parent
    assert clone._configs == obj._configs
    assert clone._configs is not obj._configs


def test_calculate_features(repo):
    groupby = ["simulation_id", "circuit_id", "neuron_class", "window"]
    features_configs_key = test_module.FeaturesConfigKey(
        groupby=groupby,
        neuron_classes=[],
        windows=[],
    )
    features_configs_list = [
        FeaturesConfig(
            type="multi",
            groupby=groupby,
            function="blueetl.external.bnac.calculate_features.calculate_features_multi",
        ),
    ]

    result = test_module._calculate_features(repo, features_configs_key, features_configs_list)
    assert isinstance(result, list)
    assert len(result) == 1

    data = result[0]
    assert isinstance(data, dict)
    assert set(data) == {
        "by_gid",
        "by_gid_and_trial",
        "by_neuron_class",
        "by_neuron_class_and_trial",
        "histograms",
    }


def test_features_collection_init(repo, features):
    assert isinstance(features, test_module.FeaturesCollection)
    assert features.cache_manager is repo.cache_manager


def test_features_collection_calculate(repo, features, capsys):
    assert features.names == [
        "by_gid",
        "by_gid_and_trial",
        "by_neuron_class",
        "by_neuron_class_and_trial",
        "histograms",
    ]

    features.calculate()
    features.show()

    captured = capsys.readouterr()
    pattern = (
        "Features: by_gid\n.*"
        "Features: by_gid_and_trial\n.*"
        "Features: by_neuron_class\n.*"
        "Features: by_neuron_class_and_trial\n.*"
        "Features: histograms\n.*"
    )
    assert re.search(pattern, captured.out, flags=re.MULTILINE | re.DOTALL)

    filtered = features.apply_filter(repo)

    assert isinstance(filtered, test_module.FilteredFeaturesCollection)


def test_features_collection_calculate_with_suffixes(repo, features_with_suffixes):
    assert features_with_suffixes.names == [
        "by_gid",
        "by_gid_0",
        "by_gid_1",
        "by_gid_and_trial",
        "by_gid_and_trial_0",
        "by_gid_and_trial_1",
        "by_neuron_class",
        "by_neuron_class_0",
        "by_neuron_class_1",
        "by_neuron_class_and_trial",
        "by_neuron_class_and_trial_0",
        "by_neuron_class_and_trial_1",
        "histograms",
        "histograms_0",
        "histograms_1",
    ]

    features_with_suffixes.calculate()

    assert isinstance(features_with_suffixes.by_gid, test_module.ConcatenatedFeatures)
    assert isinstance(features_with_suffixes.by_gid_0, Feature)
    assert isinstance(features_with_suffixes.by_gid_1, Feature)

    expected_index = [
        "simulation_id",
        "circuit_id",
        "neuron_class",
        "window",
        "gid",
        "neuron_class_index",
    ]
    expected_columns = [
        "first_spike_time_means_cort_zeroed",
        "mean_spike_counts",
        "mean_firing_rates_per_second",
    ]
    expected_extra_columns = [
        "params_id",
        "export_all_neurons",
    ]
    expected_df_0 = pd.DataFrame(
        [
            [0, 0, "L2_X", "w1", 0, 0, 0.2, 1.0, 1000.0],
            [0, 0, "L6_Y", "w1", 1, 0, 0.3, 1.0, 1000.0],
            [0, 0, "L6_Y", "w1", 2, 1, 0.1, 2.0, 2000.0],
        ],
        columns=expected_index + expected_columns,
    ).set_index(expected_index)
    expected_df_1 = expected_df_0.copy(deep=True)
    expected_df = pd.concat(
        {
            (0, True): expected_df_0,
            (1, False): expected_df_1,
        },
        names=["params_id", "export_all_neurons"],
    ).reset_index(["params_id", "export_all_neurons"])[expected_columns + expected_extra_columns]

    expected_df_0 = ensure_dtypes(expected_df_0)
    expected_df_1 = ensure_dtypes(expected_df_1)
    expected_df = ensure_dtypes(expected_df)

    expected_df_0.attrs["config"] = features_with_suffixes._features_configs[0].dict()
    expected_df_1.attrs["config"] = features_with_suffixes._features_configs[1].dict()

    assert_frame_equal(features_with_suffixes.by_gid_0.df, expected_df_0)
    assert_frame_equal(features_with_suffixes.by_gid_1.df, expected_df_1)
    assert_frame_equal(features_with_suffixes.by_gid.df, expected_df)



"""Analysis Configuration Models."""

import json
import warnings
from enum import Enum
from pathlib import Path
from typing import Annotated, Any, Optional, TypeVar, Union

from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field
from pydantic.functional_validators import field_validator, model_validator

from blueetl.constants import CONFIG_VERSION, MIN_SUPPORTED_CONFIG_VERSION
from blueetl.utils import checksum_str, dump_yaml, load_yaml

BaseModelT = TypeVar("BaseModelT", bound="BaseModel")


class BaseModel(PydanticBaseModel):
    """Custom BaseModel."""

    model_config = {
        "extra": "forbid",
        "allow_inf_nan": False,
        "validate_assignment": True,
    }

    def dict(self, *args, by_alias=True, **kwargs):
        """Generate a dictionary representation of the model, using by_alias=True by default."""
        return super().model_dump(*args, by_alias=by_alias, **kwargs)

    def json(self, *args, sort_keys=False, **kwargs):
        """Generate a JSON representation of the model, using by_alias=True by default."""
        # use json.dumps because model_dump_json in pydantic v2 doesn't support sort_keys
        return json.dumps(self.dict(*args, **kwargs), sort_keys=sort_keys, default=str)

    def dump(self, path: Path) -> None:
        """Dump the model to file in yaml format."""
        dump_yaml(path, self)

    @classmethod
    def load(cls: type[BaseModelT], path: Path) -> BaseModelT:
        """Load the model from file in yaml format."""
        return cls.model_validate(load_yaml(path))

    def checksum(self):
        """Calculate the checksum of the model."""
        return checksum_str(self.json(sort_keys=True))


class StoreTypeEnum(str, Enum):
    """StoreType Enumeration."""

    parquet = "parquet"
    feather = "feather"


class CacheConfig(BaseModel):
    """CacheConfig Model."""

    path: Path
    clear: bool = False
    readonly: bool = False
    skip_features: bool = False
    store_type: StoreTypeEnum = StoreTypeEnum.parquet

    @model_validator(mode="after")
    def validate_values(self):
        """Validate the values after loading them."""
        if self.clear and self.readonly:
            raise ValueError("clear and readonly cannot be both True at the same time")
        return self


class ReportConfig(BaseModel):
    """ReportConfig Model."""

    type: str
    name: str = ""


class WindowConfig(BaseModel):
    """WindowConfig Model."""

    initial_offset: float = 0.0
    bounds: tuple[float, float]
    t_step: float = 0.0
    n_trials: int = 0
    trial_steps_value: float = 0.0
    trial_steps_list: list[float] = []
    trial_steps_label: str = ""
    window_type: str = ""

    @model_validator(mode="after")
    def validate_values(self):
        """Validate the values after loading them."""
        if self.trial_steps_list and (self.n_trials or self.trial_steps_value):
            raise ValueError("trial_steps_list cannot be set with n_trials or trial_steps_value")
        if self.n_trials > 1 and not self.trial_steps_value:
            raise ValueError("trial_steps_value cannot be 0 when n_trials > 1")
        return self


class TrialStepsConfig(BaseModel):
    """TrialStepsConfig Model."""

    model_config = {
        **BaseModel.model_config,
        "extra": "allow",
    }
    _forbidden_extra_fields: set[str] = {
        "initial_offset",
    }
    function: str
    bounds: tuple[float, float]
    population: Optional[str] = None
    node_set: Optional[str] = None
    node_sets_file: Optional[Path] = None
    node_sets_checksum: Optional[str] = None  # to invalidate the cache when the file changes
    limit: Optional[int] = None
    base_path: Optional[Path] = None  # can be used in the function calculating the trial steps

    @model_validator(mode="after")
    def forbid_fields(self):
        """Verify that the forbidden extra fields have not been specified."""
        if found := self._forbidden_extra_fields.intersection(self.model_extra):
            raise ValueError(f"Forbidden extra fields: {found}")
        return self


class NeuronClassConfig(BaseModel):
    """NeuronClassConfig Model."""

    query: Union[dict[str, Any], list[dict[str, Any]]] = {}
    population: Optional[str] = None
    node_set: Optional[str] = None
    node_sets_file: Optional[Path] = None
    node_sets_checksum: Optional[str] = None  # to invalidate the cache when the file changes
    limit: Optional[int] = None
    node_id: Optional[list[int]] = None


class ExtractionConfig(BaseModel):
    """ExtractionConfig Model."""

    report: ReportConfig
    neuron_classes: dict[str, NeuronClassConfig] = {}
    windows: dict[str, Union[str, WindowConfig]] = {}
    trial_steps: dict[str, TrialStepsConfig] = {}

    @model_validator(mode="before")
    @classmethod
    def propagate_global_values(cls, values):
        """Propagate global values to each dictionary in neuron_classes and trial_steps."""
        for key in ["population", "node_set", "node_sets_file", "limit"]:
            if key in values:
                value = values.pop(key)
                for inner_key in ["neuron_classes", "trial_steps"]:
                    if inner_key in values:
                        for inner in values[inner_key].values():
                            inner.setdefault(key, value)
        return values


class FeaturesConfig(BaseModel):
    """FeaturesConfig Model."""

    id: Annotated[Optional[int], Field(exclude=True)] = None  # do not consider in the checksum
    type: str
    name: Optional[str] = None
    groupby: list[str]
    function: str
    neuron_classes: list[str] = []
    windows: list[str] = []
    params: dict[str, Any] = {}
    params_product: dict[str, Any] = {}
    params_zip: dict[str, Any] = {}
    suffix: str = ""
    multi_index: bool = True


class SingleAnalysisConfig(BaseModel):
    """SingleAnalysisConfig Model."""

    simulations_filter: dict[str, Any] = {}
    simulations_filter_in_memory: dict[str, Any] = {}
    extraction: ExtractionConfig
    features: list[FeaturesConfig] = []
    custom: Annotated[dict[str, Any], Field(exclude=True)] = {}  # do not dump to file

    @field_validator("features")
    @classmethod
    def assign_features_config_id(cls, lst):
        """Assign an incremental id to each FeaturesConfig."""
        for i, item in enumerate(lst):
            item.id = i
        return lst

    @model_validator(mode="before")
    @classmethod
    def handle_deprecated_fields(cls, data: Any) -> Any:
        """Handle the deprecated fields."""
        # needed to load any cached analysis configuration created with blueetl <= 0.8.2
        data.pop("output", None)
        return data


class MultiAnalysisConfig(BaseModel):
    """MultiAnalysisConfig Model."""

    version: int
    simulation_campaign: Path
    cache: CacheConfig
    simulations_filter: dict[str, Any] = {}
    simulations_filter_in_memory: dict[str, Any] = {}
    analysis: dict[str, SingleAnalysisConfig]
    custom: Annotated[dict[str, Any], Field(exclude=True)] = {}  # do not dump to file

    @field_validator("version")
    @classmethod
    def version_match(cls, version):
        """Verify that the config version is supported."""
        if version < MIN_SUPPORTED_CONFIG_VERSION:
            raise ValueError(f"Config version {version} is not supported.")
        if version != CONFIG_VERSION:
            warnings.warn(f"version {version} is deprecated, see the docs to update the config")
        return version

    @model_validator(mode="before")
    @classmethod
    def handle_deprecated_fields(cls, data: Any) -> Any:
        """Handle the deprecated fields."""
        cache_config = data.setdefault("cache", {})
        if (value := data.pop("output", None)) is not None:
            if "path" in cache_config:
                raise ValueError("output must be removed when path is defined")
            warnings.warn("output is deprecated, use cache.path instead")
            cache_config["path"] = value
        if (value := data.pop("clear_cache", None)) is not None:
            if "clear_cache" in cache_config:
                raise ValueError("clear_cache must be removed when clear is defined")
            warnings.warn("clear_cache is deprecated, use cache.clear instead")
            cache_config["clear"] = value
        return data


if __name__ == "__main__":
    import yaml

    # print the json schema of the model in yaml format,
    # to be used only to update the current schema
    print(yaml.safe_dump(MultiAnalysisConfig.model_json_schema(), sort_keys=False))



import os
from unittest.mock import MagicMock, Mock, PropertyMock, patch

import pandas as pd
from blueetl_core.constants import BLUEETL_JOBLIB_JOBS
from pandas.testing import assert_frame_equal

from blueetl.constants import (
    CIRCUIT,
    CIRCUIT_ID,
    COUNT,
    DURATION,
    GID,
    GIDS,
    LIMIT,
    NEURON_CLASS,
    NEURON_CLASS_INDEX,
    NODE_SET,
    OFFSET,
    POPULATION,
    QUERY,
    SIMULATION,
    SIMULATION_ID,
    T_START,
    T_STEP,
    T_STOP,
    TRIAL,
    WINDOW,
    WINDOW_TYPE,
)
from blueetl.extract import spikes as test_module
from blueetl.utils import ensure_dtypes


def _get_spikes(gids):
    """Return a Series as returned by simulation.spikes[population].get()."""
    spikes = pd.Series(
        [300, 100, 300, 200, 100, 100],
        index=pd.Index([56.05, 82.25, 441.85, 520.025, 609.425, 1167.525], name="times"),
        name="ids",
    )
    return spikes[spikes.isin(gids)]


@patch.dict(os.environ, {BLUEETL_JOBLIB_JOBS: "1"})
def test_spikes_from_simulations():
    mock_circuit = MagicMock()
    mock_sim = MagicMock()
    mock_sim.spikes.__getitem__.return_value.get.side_effect = _get_spikes
    mock_simulations_df = PropertyMock(
        return_value=pd.DataFrame(
            [
                {SIMULATION_ID: 0, CIRCUIT_ID: 0, SIMULATION: mock_sim, CIRCUIT: mock_circuit},
            ]
        )
    )
    mock_simulations = Mock()
    type(mock_simulations).df = mock_simulations_df

    mock_neurons_df = PropertyMock(
        return_value=pd.DataFrame(
            [
                {CIRCUIT_ID: 0, NEURON_CLASS: "L23_EXC", GID: 100, NEURON_CLASS_INDEX: 0},
                {CIRCUIT_ID: 0, NEURON_CLASS: "L23_EXC", GID: 200, NEURON_CLASS_INDEX: 1},
                {CIRCUIT_ID: 0, NEURON_CLASS: "L4_EXC", GID: 300, NEURON_CLASS_INDEX: 0},
            ]
        )
    )
    mock_neurons = Mock()
    type(mock_neurons).df = mock_neurons_df

    mock_windows_df = PropertyMock(
        return_value=pd.DataFrame(
            [
                {
                    SIMULATION_ID: 0,
                    CIRCUIT_ID: 0,
                    WINDOW: "w1",
                    TRIAL: 0,
                    OFFSET: 20,
                    T_START: 0,
                    T_STOP: 100,
                    T_STEP: 0,
                    DURATION: 100,
                    WINDOW_TYPE: "spontaneous",
                },
            ]
        )
    )
    mock_windows = Mock()
    type(mock_windows).df = mock_windows_df

    mock_neuron_classes_df = PropertyMock(
        return_value=pd.DataFrame(
            [
                {
                    CIRCUIT_ID: 0,
                    NEURON_CLASS: "L23_EXC",
                    COUNT: 2,
                    LIMIT: None,
                    POPULATION: "thalamus_neurons",
                    NODE_SET: None,
                    GIDS: None,
                    QUERY: "{}",
                },
                {
                    CIRCUIT_ID: 0,
                    NEURON_CLASS: "L4_EXC",
                    COUNT: 1,
                    LIMIT: None,
                    POPULATION: "thalamus_neurons",
                    NODE_SET: None,
                    GIDS: None,
                    QUERY: "{}",
                },
            ]
        )
    )
    mock_neuron_classes = Mock()
    type(mock_neuron_classes).df = mock_neuron_classes_df

    result = test_module.Spikes.from_simulations(
        simulations=mock_simulations,
        neurons=mock_neurons,
        windows=mock_windows,
        neuron_classes=mock_neuron_classes,
        name="spikes",
    )

    expected_df = pd.DataFrame(
        [
            {
                "time": 62.25,
                "gid": 100,
                "window": "w1",
                "trial": 0,
                "simulation_id": 0,
                "circuit_id": 0,
                "neuron_class": "L23_EXC",
            },
            {
                "time": 36.05,
                "gid": 300,
                "window": "w1",
                "trial": 0,
                "simulation_id": 0,
                "circuit_id": 0,
                "neuron_class": "L4_EXC",
            },
        ]
    )
    expected_df = ensure_dtypes(expected_df)
    assert isinstance(result, test_module.Spikes)
    assert_frame_equal(result.df, expected_df)
    assert mock_simulations_df.call_count == 1
    assert mock_neurons_df.call_count == 1
    assert mock_windows_df.call_count == 1



import pandas as pd
import pytest
import xarray as xr

from blueetl.campaign.config import SimulationCampaign


@pytest.fixture
def xarray_config_dict_simple():
    return {
        "dims": ["ca", "depolarization"],
        "attrs": {
            "path_prefix": "/tmp/simple",
            "blue_config_template": "simulation_sonata.tmpl",
            "circuit_config": "/tmp/tests/data/circuit_sonata.json",
        },
        "data": [["", "uuid/1"], ["uuid/2", "uuid/3"]],
        "coords": {
            "ca": {"dims": ["ca"], "attrs": {}, "data": [1.0, 2.0]},
            "depolarization": {"dims": ["depolarization"], "attrs": {}, "data": [3.0, 4.0]},
        },
        "name": "uuid",
    }


@pytest.fixture
def xarray_config_dict_coupled():
    return {
        "dims": ["coupled"],
        "attrs": {
            "path_prefix": "/tmp/coupled",
            "blue_config_template": "simulation_sonata.tmpl",
            "circuit_config": "/tmp/tests/data/circuit_sonata.json",
        },
        "data": ["", "uuid/1"],
        "coords": {
            "ca": {"dims": ["coupled"], "attrs": {}, "data": [1.0, 2.0]},
            "depolarization": {"dims": ["coupled"], "attrs": {}, "data": [3.0, 4.0]},
        },
        "name": "uuid",
    }


@pytest.fixture
def blueetl_config_dict_simple():
    return {
        "format": "blueetl",
        "version": 1,
        "name": "uuid",
        "attrs": {
            "path_prefix": "/tmp/simple",
            "blue_config_template": "simulation_sonata.tmpl",
            "circuit_config": "/tmp/tests/data/circuit_sonata.json",
        },
        "data": [
            {
                "ca": 1.0,
                "depolarization": 3.0,
                "simulation_path": "",
            },
            {
                "ca": 1.0,
                "depolarization": 4.0,
                "simulation_path": "/tmp/simple/uuid/1/simulation_config.json",
            },
            {
                "ca": 2.0,
                "depolarization": 3.0,
                "simulation_path": "/tmp/simple/uuid/2/simulation_config.json",
            },
            {
                "ca": 2.0,
                "depolarization": 4.0,
                "simulation_path": "/tmp/simple/uuid/3/simulation_config.json",
            },
        ],
    }


@pytest.fixture
def blueetl_config_dict_coupled():
    return {
        "format": "blueetl",
        "version": 1,
        "name": "uuid",
        "attrs": {
            "path_prefix": "/tmp/coupled",
            "blue_config_template": "simulation_sonata.tmpl",
            "circuit_config": "/tmp/tests/data/circuit_sonata.json",
            "__coupled__": "coupled",
        },
        "data": [
            {
                "ca": 1.0,
                "depolarization": 3.0,
                "simulation_path": "",
            },
            {
                "ca": 2.0,
                "depolarization": 4.0,
                "simulation_path": "/tmp/coupled/uuid/1/simulation_config.json",
            },
        ],
    }


@pytest.fixture
def xarray_config_obj_simple(xarray_config_dict_simple):
    return xr.DataArray.from_dict(xarray_config_dict_simple)


@pytest.fixture
def xarray_config_obj_coupled(xarray_config_dict_coupled):
    return xr.DataArray.from_dict(xarray_config_dict_coupled)


@pytest.fixture
def blueetl_config_dataframe_simple(blueetl_config_dict_simple):
    d = blueetl_config_dict_simple
    return pd.DataFrame.from_records(d["data"])


@pytest.fixture
def blueetl_config_dataframe_coupled(blueetl_config_dict_coupled):
    d = blueetl_config_dict_coupled
    return pd.DataFrame.from_records(d["data"])


@pytest.fixture
def blueetl_config_obj_simple(blueetl_config_dict_simple):
    d = blueetl_config_dict_simple
    return SimulationCampaign(
        data=pd.DataFrame.from_records(d["data"]),
        name=d["name"],
        attrs=d["attrs"],
    )


@pytest.fixture
def blueetl_config_obj_coupled(blueetl_config_dict_coupled):
    d = blueetl_config_dict_coupled
    return SimulationCampaign(
        data=pd.DataFrame.from_records(d["data"]),
        name=d["name"],
        attrs=d["attrs"],
    )



"""Base Adapter."""

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Generic, Optional, TypeVar

from blueetl.adapters.interfaces.circuit import CircuitInterface
from blueetl.adapters.interfaces.node_sets import NodeSetsInterface
from blueetl.adapters.interfaces.simulation import SimulationInterface

InterfaceT = TypeVar("InterfaceT", CircuitInterface, SimulationInterface, NodeSetsInterface)
BaseAdapterT = TypeVar("BaseAdapterT", bound="BaseAdapter")


class AdapterError(Exception):
    """Generic Adapter Error."""


class BaseAdapter(Generic[InterfaceT], ABC):
    """Base Adapter to be subclassed by other adapters."""

    def __init__(self, _impl: Optional[InterfaceT]) -> None:
        """Init the adapter from the specified implementation."""
        self._impl: Optional[InterfaceT] = _impl

    @classmethod
    @abstractmethod
    def from_file(cls, filepath: Optional[Path]) -> "BaseAdapter":
        """Load and return a new object from file."""

    @property
    def _ensure_impl(self) -> InterfaceT:
        """Return the inner implementation, or raise an error if it doesn't exist."""
        if self._impl is None:
            raise AdapterError("The implementation doesn't exist")
        return self._impl

    def exists(self) -> bool:
        """Return True if the wrapped object exists, False otherwise."""
        return self._impl is not None

    @property
    def instance(self) -> Any:
        """Return the wrapped instance, or None if it doesn't exist."""
        return self._impl.instance if self._impl is not None else None



"""Spikes extractor."""

import logging
from typing import Optional

import pandas as pd
from blueetl_core.utils import smart_concat

from blueetl.adapters.simulation import SimulationAdapter as Simulation
from blueetl.constants import CIRCUIT_ID, GID, NEURON_CLASS, SIMULATION_ID, TIME, TRIAL, WINDOW
from blueetl.extract.report import ReportExtractor

L = logging.getLogger(__name__)


class Spikes(ReportExtractor):
    """Spikes extractor class."""

    COLUMNS = [SIMULATION_ID, CIRCUIT_ID, NEURON_CLASS, WINDOW, TRIAL, TIME, GID]

    @classmethod
    def _assign_window(cls, df: pd.DataFrame, rec) -> pd.DataFrame:
        win = cls.calculate_window_slice(rec)
        df = df[(df[TIME] >= win.t_start) & (df[TIME] < win.t_stop)].copy()
        df[WINDOW] = win.name
        df[TRIAL] = win.trial
        # make the spike times relative to the offset
        df[TIME] -= win.offset
        return df

    @classmethod
    def _load_values(
        cls,
        simulation: Simulation,
        population: Optional[str],
        gids,
        windows_df: pd.DataFrame,
        name: str,
    ) -> pd.DataFrame:
        """Filter and aggregate the spikes in bins according to the given windows.

        Args:
            simulation: simulation containing the SpikeReport of times and gids.
            population: node population name.
            gids: array of gids to be selected.
            windows_df: windows dataframe with columns [window, trial, t_start, t_stop]
            name: name of the report in the simulation configuration, ignored.

        Returns:
            pd.DataFrame: dataframe with columns [window, time, gid]
        """
        df = simulation.spikes[population].get(gids).reset_index()
        # in snap the columns are named `times` and `ids`
        df.columns.array[0:2] = [TIME, GID]
        df = smart_concat(cls._assign_window(df, rec) for rec in windows_df.itertuples())
        df = df.reset_index(drop=True)
        return df



import pickle
from pathlib import Path

import pytest

from blueetl.adapters import circuit as test_module
from blueetl.adapters.base import AdapterError
from blueetl.adapters.node_sets import NodeSetsAdapter
from tests.unit.utils import BLUEPY_AVAILABLE, TEST_DATA_PATH, assert_isinstance


@pytest.mark.parametrize(
    "path, population, expected_classes",
    [
        (
            pytest.param(
                "sonata/circuit_config.json",
                "default",
                {
                    "circuit": "bluepysnap.Circuit",
                    "population": "bluepysnap.nodes.NodePopulation",
                },
                id="snap",
            )
        ),
        (
            pytest.param(
                "bbp/CircuitConfig",
                None,
                {
                    "circuit": "bluepy.Circuit",
                    "population": "bluepy.cells.CellCollection",
                },
                id="bluepy",
                marks=pytest.mark.skipif(not BLUEPY_AVAILABLE, reason="bluepy not available"),
            )
        ),
    ],
)
def test_circuit_adapter(path, population, expected_classes, monkeypatch):
    path = TEST_DATA_PATH / "circuit" / path
    # enter the circuit dir to resolve relative paths in bluepy
    monkeypatch.chdir(path.parent)
    obj = test_module.CircuitAdapter.from_file(path)
    assert_isinstance(obj.instance, expected_classes["circuit"])

    # access methods and properties
    pop = obj.nodes[population]
    assert_isinstance(pop, expected_classes["population"])

    checksum = obj.checksum()
    assert isinstance(checksum, str)

    # test pickle roundtrip
    dumped = pickle.dumps(obj)
    loaded = pickle.loads(dumped)

    assert isinstance(loaded, test_module.CircuitAdapter)
    assert_isinstance(loaded.instance, expected_classes["circuit"])
    # no cached_properties should be loaded after unpickling
    assert sorted(loaded.__dict__) == ["_impl"]
    assert sorted(loaded._impl.__dict__) == ["_circuit"]


def test_circuit_adapter_with_nonexistent_path():
    path = Path("path/to/circuit_config.json")
    obj = test_module.CircuitAdapter.from_file(path)

    assert obj.instance is None
    assert obj.exists() is False

    with pytest.raises(AdapterError, match="The implementation doesn't exist"):
        _ = obj.nodes

    with pytest.raises(AdapterError, match="The implementation doesn't exist"):
        _ = obj.checksum()

    # test pickle roundtrip
    dumped = pickle.dumps(obj)
    loaded = pickle.loads(dumped)

    assert isinstance(loaded, test_module.CircuitAdapter)
    assert loaded.instance is None


def test_circuit_adapter_node_sets():
    path = TEST_DATA_PATH / "circuit" / "sonata" / "circuit_config.json"
    obj = test_module.CircuitAdapter.from_file(path)

    assert_isinstance(obj.instance, "bluepysnap.Circuit")

    assert isinstance(obj.node_sets_file, Path)
    assert obj.node_sets_file.name == "node_sets.json"

    assert isinstance(obj.node_sets, NodeSetsAdapter)



"""Adapters for NodeSets."""

from pathlib import Path
from typing import Optional

from blueetl.adapters.base import BaseAdapter
from blueetl.adapters.interfaces.node_sets import NodeSetsInterface


class NodeSetsAdapter(BaseAdapter[NodeSetsInterface]):
    """NodeSets Adapter."""

    @classmethod
    def from_file(cls, filepath: Optional[Path]) -> "NodeSetsAdapter":
        """Load and return a new object from file."""
        # pylint: disable=import-outside-toplevel
        from blueetl.adapters.impl.bluepysnap.node_sets import NodeSets, NodeSetsImpl

        if not filepath:
            impl = NodeSetsImpl(NodeSets.from_dict({}))
        else:
            impl = NodeSetsImpl(NodeSets.from_file(str(filepath)))
        return cls(impl)

    def update(self, other: "NodeSetsAdapter") -> None:
        """Update the node sets."""
        # pylint: disable=protected-access
        self._ensure_impl.update(other._ensure_impl)

    def __ior__(self, other: "NodeSetsAdapter") -> "NodeSetsAdapter":
        """Support ``A |= B``."""
        self.update(other)
        return self



import pytest

from blueetl import validation as test_module


def test_read_schema():
    schema = test_module.read_schema("analysis_config")
    assert isinstance(schema, dict)


@pytest.mark.parametrize(
    "analysis_config",
    [
        {
            "version": 3,
            "simulation_campaign": "/path/to/simulation/campaign",
            "output": "/path/to/output",
            "analysis": {},
        },
        {
            "version": 4,
            "simulation_campaign": "/path/to/simulation/campaign",
            "cache": {"path": "/path/to/output"},
            "analysis": {},
        },
        {
            "version": 3,
            "simulation_campaign": "/path/to/simulation/campaign",
            "output": "/path/to/output",
            "analysis": {
                "spikes": {
                    "extraction": {
                        "report": {"type": "spikes"},
                        "population": "default",
                        "neuron_classes": {},
                        "windows": {},
                    }
                },
            },
        },
    ],
)
def test_validate_config(analysis_config):
    schema = test_module.read_schema("analysis_config")
    test_module.validate_config(config=analysis_config, schema=schema)


@pytest.mark.parametrize(
    "analysis_config",
    [
        {},
        {
            "version": 3,
            "simulation_campaign": "/path/to/simulation/campaign",
            "output": "/path/to/output",
            "analysis": {
                "spikes": {"extraction": {}},
            },
        },
    ],
)
def test_validate_config_invalid(analysis_config):
    schema = test_module.read_schema("analysis_config")
    with pytest.raises(test_module.ValidationError):
        test_module.validate_config(config=analysis_config, schema=schema)



from pathlib import Path

import numpy as np
import pandas as pd
from bluepysnap import Simulation
from pandas.testing import assert_frame_equal, assert_series_equal

from blueetl.campaign.config import SimulationCampaign
from blueetl.converters import convert_spikes as test_module


def test_main(tmp_path):
    population = "custom"
    input_file = tmp_path / "spikes.csv"
    output_dir = tmp_path / "output"
    input_file.write_text("ids times\n1 10.0\n1 10.1\n2 10.1\n2 20.0\n5 5.0")
    test_module.main(input_file, output_dir, node_population=population)

    assert (output_dir / "analysis_config.yaml").is_file()
    assert (output_dir / "circuit" / "circuit_config.json").is_file()
    assert (output_dir / "circuit" / "node_sets.json").is_file()
    assert (output_dir / "circuit" / "synthetic_nodes.h5").is_file()
    assert (output_dir / "simulation" / "simulation_config.json").is_file()
    assert (output_dir / "simulation" / "synthetic_spikes.h5").is_file()
    assert (output_dir / "simulation_campaign_config.json").is_file()

    campaign = SimulationCampaign.load(output_dir / "simulation_campaign_config.json")
    assert len(campaign) == 1
    simulation = Simulation(campaign[0].path)
    assert (
        Path(simulation._simulation_config_path).resolve()
        == (output_dir / "simulation" / "simulation_config.json").resolve()
    )
    assert (
        Path(simulation.circuit._circuit_config_path).resolve()
        == (output_dir / "circuit" / "circuit_config.json").resolve()
    )

    expected_nodes = pd.DataFrame(
        {"_": np.zeros(6, dtype=np.int8)},
        index=pd.RangeIndex(6, name="node_ids"),
    )
    expected_spikes = pd.Series(
        [5, 1, 1, 2, 2],
        index=pd.Index([5.0, 10.0, 10.1, 10.1, 20.0], name="times"),
        name="ids",
    )
    assert_frame_equal(simulation.circuit.nodes[population].get(), expected_nodes)
    assert_series_equal(simulation.spikes[population].get(), expected_spikes)



import importlib
from functools import partial
from pathlib import Path
from unittest.mock import MagicMock

import pandas as pd
import pydantic
import xarray as xr

try:
    import bluepy

    BLUEPY_AVAILABLE = True
except ImportError:
    BLUEPY_AVAILABLE = False

TEST_DATA_PATH = Path(__file__).parent / "data"
TEST_NODE_SETS_FILE = TEST_DATA_PATH / "circuit" / "sonata" / "node_sets.json"
TEST_NODE_SETS_FILE_EXTRA = TEST_DATA_PATH / "circuit" / "sonata" / "node_sets_extra.json"
TEST_CIRCUIT_CONFIG = TEST_DATA_PATH / "circuit" / "sonata" / "circuit_config.json"
TEST_SIMULATION_CONFIG = TEST_DATA_PATH / "simulation" / "sonata" / "simulation_config.json"


class PicklableMock(MagicMock):
    """Mock that can be pickled without errors.

    Neither the internal status nor any attribute set after the initialization is saved or restored!
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._args = args
        self._kwargs = kwargs

    __reduce__ = lambda self: (partial(MagicMock, **self._kwargs), self._args)


def assert_frame_equal(actual, expected):
    pd.testing.assert_frame_equal(actual, expected)
    # check explicitly attrs because not considered by pd.testing.assert_frame_equal
    assert actual.attrs == expected.attrs


def assert_xr_equal(actual, expected):
    xr.testing.assert_equal(actual, expected)
    # check explicitly name and attrs because not considered by xr.testing.assert_equal
    assert actual.name == expected.name
    assert actual.attrs == expected.attrs


def iterallvalues(obj):
    if isinstance(obj, dict):
        for v in obj.values():
            yield from iterallvalues(v)
    elif isinstance(obj, (list, tuple)):
        for v in obj:
            yield from iterallvalues(v)
    elif isinstance(obj, pydantic.BaseModel):
        for k, v in obj:
            yield from iterallvalues(v)
    yield obj


def assert_not_duplicates(obj):
    """Verify that obj doesn't contain duplicate instances."""
    ids = set()
    for v in iterallvalues(obj):
        if isinstance(v, (dict, list, tuple)):
            assert id(v) not in ids, f"Duplicate {type(v).__name__}: {v}"
            ids.add(id(v))


def assert_isinstance(instance, class_name):
    """Verify that instance is an instance of class_name (given as string)."""
    module_name, _, class_name = class_name.rpartition(".")
    cls = getattr(importlib.import_module(module_name), class_name)
    assert isinstance(instance, cls)



"""Trial steps functions adapted from BlueNetworkActivityComparison/bnac/onset.py."""

import logging
from pathlib import Path

import numpy as np
from scipy.ndimage import gaussian_filter

L = logging.getLogger(__name__)


def _get_bounds(params):
    lower_bound, upper_bound = params["bounds"]
    assert lower_bound <= 0
    assert upper_bound >= 0
    return lower_bound, upper_bound


def _histogram_from_spikes(spikes, params):
    lower_bound, upper_bound = _get_bounds(params)
    onset_test_window_length = upper_bound - lower_bound
    histogram, _ = np.histogram(
        spikes,
        range=(lower_bound, upper_bound),
        bins=int(onset_test_window_length * params["histo_bins_per_ms"]),
    )
    return histogram


def _onset_from_histogram(histogram, params):
    lower_bound, _ = _get_bounds(params)
    smoothed_histogram = gaussian_filter(histogram, sigma=params["smoothing_width"])

    pre_window_length = -lower_bound
    pre_window_bins = int(pre_window_length * params["histo_bins_per_ms"])
    pre_smoothed_histogram = smoothed_histogram[:pre_window_bins]
    post_smoothed_histogram = smoothed_histogram[pre_window_bins:]

    onset_dict = {
        "pre_mean": np.mean(pre_smoothed_histogram),
        "pre_std": np.std(pre_smoothed_histogram),
        "post_max": np.max(post_smoothed_histogram),
    }
    onset_dict["pre_mean_post_max_ratio"] = onset_dict["pre_mean"] / onset_dict["post_max"]

    threshold = onset_dict["pre_mean"] + params["threshold_std_multiple"] * onset_dict["pre_std"]
    where_above_thresh = np.where(post_smoothed_histogram > threshold)[0]
    index_above_std = 0
    if len(where_above_thresh) > 0:
        index_above_std = where_above_thresh[0]
    # index_above_count_thresh = np.where(post_smoothed_histogram > 15)[0][0]
    # index_above_count_thresh = -1000
    # cortical_onset_index = np.max([index_above_std, index_above_count_thresh])
    cortical_onset_index = index_above_std

    onset_dict["cortical_onset"] = (
        float(cortical_onset_index) / float(params["histo_bins_per_ms"]) + params["ms_post_offset"]
    )
    if params.get("figures_path"):
        _plot(smoothed_histogram, params, onset_dict)
    return onset_dict


def _plot(smoothed_histogram, params, onset_dict):
    # pylint: disable=import-outside-toplevel
    import matplotlib

    matplotlib.use("agg")
    import matplotlib.pyplot as plt
    import seaborn as sns

    lower_bound, upper_bound = _get_bounds(params)
    plt.figure()
    sns.set()
    sns.set_style("ticks")
    x_vals = list(np.arange(lower_bound, upper_bound, 0.2))
    plt.plot(x_vals, smoothed_histogram)
    plt.scatter(
        onset_dict["cortical_onset"],
        [onset_dict["pre_mean"] + params["threshold_std_multiple"] * onset_dict["pre_std"]],
    )
    plt.gca().set_xlim([lower_bound, upper_bound])
    plt.gca().set_xlabel("Time (ms)")
    plt.gca().set_ylabel("Number of spikes")
    plt.gca().spines["right"].set_visible(False)
    plt.gca().spines["top"].set_visible(False)
    filepath = Path(params["base_path"], params["figures_path"], "plot.pdf")
    L.info("Figures path: %s", filepath)
    filepath.parent.mkdir(exist_ok=True)
    plt.savefig(filepath)
    plt.close()


def onset_from_spikes(spikes_list, params):
    """Calculate the cortical onset from a list of spikes, one for each trial.

    Args:
        spikes_list: list of spikes as numpy arrays.
        params: dictionary of parameters from the trial steps configuration.

    Returns:
        float representing the dynamic offset to be added to the initial offset of each trial step.
    """
    L.info(
        "onset_from_spikes: processing %s arrays of spikes using params=%r",
        len(spikes_list),
        params,
    )
    spikes = np.concatenate(spikes_list)
    histogram = _histogram_from_spikes(spikes, params)
    onset_dict = _onset_from_histogram(histogram, params)
    return onset_dict["cortical_onset"]



import os
from unittest.mock import MagicMock, Mock, PropertyMock, patch

import pandas as pd
from blueetl_core.constants import BLUEETL_JOBLIB_JOBS
from pandas.testing import assert_frame_equal

from blueetl.constants import (
    CIRCUIT,
    CIRCUIT_ID,
    COUNT,
    DURATION,
    GID,
    GIDS,
    LIMIT,
    NEURON_CLASS,
    NEURON_CLASS_INDEX,
    NODE_SET,
    OFFSET,
    POPULATION,
    QUERY,
    SIMULATION,
    SIMULATION_ID,
    T_START,
    T_STEP,
    T_STOP,
    TRIAL,
    WINDOW,
    WINDOW_TYPE,
)
from blueetl.extract import soma_report as test_module
from blueetl.utils import ensure_dtypes


def _get_soma_report(group=None, t_start=None, t_stop=None, t_step=None):
    """Return a DataFrame as returned by simulation.reports["soma"]["pop"].get()."""
    df = pd.DataFrame(
        {
            100: [-72.1, -72.3, -72.5],
            200: [-70.7, -70.8, -71.0],
            300: [-69.5, -69.8, -70.0],
        },
        index=pd.Index([19.5, 20, 20.5], name="time"),
    )
    return df.etl.q(time={"ge": t_start, "lt": t_stop})[list(group)]


@patch.dict(os.environ, {BLUEETL_JOBLIB_JOBS: "1"})
def test_soma_report_from_simulations():
    mock_circuit = MagicMock()
    mock_sim = MagicMock()
    _report_by_type = mock_sim.reports.__getitem__.return_value
    _report_by_pop = _report_by_type.__getitem__.return_value
    _report_by_pop.get.side_effect = _get_soma_report
    mock_simulations_df = PropertyMock(
        return_value=pd.DataFrame(
            [
                {SIMULATION_ID: 0, CIRCUIT_ID: 0, SIMULATION: mock_sim, CIRCUIT: mock_circuit},
            ]
        )
    )
    mock_simulations = Mock()
    type(mock_simulations).df = mock_simulations_df

    mock_neurons_df = PropertyMock(
        return_value=pd.DataFrame(
            [
                {CIRCUIT_ID: 0, NEURON_CLASS: "L23_EXC", GID: 100, NEURON_CLASS_INDEX: 0},
                {CIRCUIT_ID: 0, NEURON_CLASS: "L23_EXC", GID: 200, NEURON_CLASS_INDEX: 1},
                {CIRCUIT_ID: 0, NEURON_CLASS: "L4_EXC", GID: 300, NEURON_CLASS_INDEX: 0},
            ]
        )
    )
    mock_neurons = Mock()
    type(mock_neurons).df = mock_neurons_df

    mock_windows_df = PropertyMock(
        return_value=pd.DataFrame(
            [
                {
                    SIMULATION_ID: 0,
                    CIRCUIT_ID: 0,
                    WINDOW: "w1",
                    TRIAL: 0,
                    OFFSET: 20,
                    T_START: 0,
                    T_STOP: 100,
                    T_STEP: 0,
                    DURATION: 100,
                    WINDOW_TYPE: "spontaneous",
                },
            ]
        )
    )
    mock_windows = Mock()
    type(mock_windows).df = mock_windows_df

    mock_neuron_classes_df = PropertyMock(
        return_value=pd.DataFrame(
            [
                {
                    CIRCUIT_ID: 0,
                    NEURON_CLASS: "L23_EXC",
                    COUNT: 2,
                    LIMIT: None,
                    POPULATION: "thalamus_neurons",
                    NODE_SET: None,
                    GIDS: None,
                    QUERY: "{}",
                },
                {
                    CIRCUIT_ID: 0,
                    NEURON_CLASS: "L4_EXC",
                    COUNT: 1,
                    LIMIT: None,
                    POPULATION: "thalamus_neurons",
                    NODE_SET: None,
                    GIDS: None,
                    QUERY: "{}",
                },
            ]
        )
    )
    mock_neuron_classes = Mock()
    type(mock_neuron_classes).df = mock_neuron_classes_df

    result = test_module.SomaReport.from_simulations(
        simulations=mock_simulations,
        neurons=mock_neurons,
        windows=mock_windows,
        neuron_classes=mock_neuron_classes,
        name="soma",
    )

    expected_df = pd.DataFrame(
        [
            {
                "gid": 100,
                "time": 20.0,
                "value": -72.3,
                "window": "w1",
                "simulation_id": 0,
                "circuit_id": 0,
                "neuron_class": "L23_EXC",
            },
            {
                "gid": 100,
                "time": 20.5,
                "value": -72.5,
                "window": "w1",
                "simulation_id": 0,
                "circuit_id": 0,
                "neuron_class": "L23_EXC",
            },
            {
                "gid": 200,
                "time": 20.0,
                "value": -70.8,
                "window": "w1",
                "simulation_id": 0,
                "circuit_id": 0,
                "neuron_class": "L23_EXC",
            },
            {
                "gid": 200,
                "time": 20.5,
                "value": -71.0,
                "window": "w1",
                "simulation_id": 0,
                "circuit_id": 0,
                "neuron_class": "L23_EXC",
            },
            {
                "gid": 300,
                "time": 20.0,
                "value": -69.8,
                "window": "w1",
                "simulation_id": 0,
                "circuit_id": 0,
                "neuron_class": "L4_EXC",
            },
            {
                "gid": 300,
                "time": 20.5,
                "value": -70.0,
                "window": "w1",
                "simulation_id": 0,
                "circuit_id": 0,
                "neuron_class": "L4_EXC",
            },
        ]
    )
    expected_df = ensure_dtypes(expected_df)
    assert isinstance(result, test_module.SomaReport)
    assert_frame_equal(result.df, expected_df)
    assert mock_simulations_df.call_count == 1
    assert mock_neurons_df.call_count == 1
    assert mock_windows_df.call_count == 1



"""Bluepysnap circuit implementation."""

from collections.abc import Mapping
from pathlib import Path
from typing import Optional

from bluepysnap import Circuit

from blueetl.adapters.impl.bluepysnap.node_sets import NodeSetsImpl
from blueetl.adapters.interfaces.circuit import CircuitInterface, NodePopulationInterface
from blueetl.adapters.interfaces.node_sets import NodeSetsInterface
from blueetl.utils import checksum_json


class CircuitImpl(CircuitInterface[Circuit]):
    """Bluepysnap circuit implementation."""

    def checksum(self) -> str:
        """Return a hash of the relevant keys in the circuit configuration.

        Circuits are considered different in this context if any relevant key is different.
        """
        return checksum_json(self._circuit.config)

    @property
    def nodes(self) -> Mapping[Optional[str], NodePopulationInterface]:
        """Return the nodes as a dict: population -> nodes."""
        return self._circuit.nodes

    @property
    def node_sets_file(self) -> Optional[Path]:
        """Returns the NodeSets file used by the circuit."""
        path = self._circuit.to_libsonata.node_sets_path
        return Path(path) if path else None

    @property
    def node_sets(self) -> NodeSetsInterface:
        """Returns the NodeSets used by the circuit."""
        return NodeSetsImpl(self._circuit.node_sets)



import json
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import yaml
from numpy.testing import assert_equal
from packaging.version import Version
from pandas.api.types import is_integer_dtype

from blueetl import utils as test_module
from blueetl.constants import DTYPES
from tests.unit.utils import TEST_DATA_PATH


@pytest.mark.parametrize(
    "x, expected",
    [
        ([], []),
        ((), []),
        (None, [None]),
        ("", [""]),
        ("a", ["a"]),
        (["a"], ["a"]),
        (("a",), ["a"]),
        ({"a": 1}, [{"a": 1}]),
    ],
)
def test_ensure_list(x, expected):
    result = test_module.ensure_list(x)
    assert result == expected


@pytest.mark.parametrize("dtypes", [None, DTYPES])
def test_ensure_dtypes(dtypes):
    df = pd.DataFrame([{k: i for i, k in enumerate(DTYPES)}])
    df["other"] = 99
    df.index = pd.MultiIndex.from_frame(df)

    result = test_module.ensure_dtypes(df, desired_dtypes=dtypes)

    # verify the dtypes of the columns
    expected = [*list(DTYPES.values()), np.int64]
    assert_equal(result.dtypes.to_numpy(), expected)

    # verify the dtypes of the MultiIndex
    # Changed in version 2.0.0: Index can hold all numpy numeric dtypes (except float16).
    # Previously only int64/uint64/float64 dtypes were accepted.
    if Version(pd.__version__) < Version("2"):
        # convert any int to int64 in the expected list
        expected = [np.int64 if is_integer_dtype(item) else item for item in expected]
    assert_equal(result.index.dtypes.to_numpy(), expected)


def test_import_by_string():
    name = "json.load"
    result = test_module.import_by_string(name)
    assert result is json.load


def test_resolve_path(tmp_path):
    subdir = Path("subdir")
    my_file = Path("my_file")
    my_symlink = Path("my_symlink")
    non_existent = Path("non_existent")

    tmp_path = tmp_path.resolve()
    subdir_absolute = tmp_path / subdir
    my_file_absolute = tmp_path / subdir / my_file
    my_symlink_absolute = tmp_path / subdir / my_symlink
    non_existent_absolute = tmp_path / subdir / non_existent

    subdir_absolute.mkdir()
    my_file_absolute.touch()
    my_symlink_absolute.symlink_to(my_file)

    result = test_module.resolve_path(tmp_path, subdir, my_file, symlinks=False)
    assert result == my_file_absolute

    result = test_module.resolve_path(tmp_path, subdir, my_file, symlinks=True)
    assert result == my_file_absolute

    result = test_module.resolve_path(tmp_path, subdir, my_symlink, symlinks=False)
    assert result == my_symlink_absolute

    result = test_module.resolve_path(tmp_path, subdir, my_symlink, symlinks=True)
    assert result == my_file_absolute

    result = test_module.resolve_path(tmp_path, subdir, non_existent, symlinks=False)
    assert result == non_existent_absolute

    result = test_module.resolve_path(tmp_path, subdir, non_existent, symlinks=True)
    assert result == non_existent_absolute


def test_dump_yaml(tmp_path):
    data = {
        "dict": {"str": "mystr", "int": 123},
        "list_of_int": [1, 2, 3],
        "list_of_str": ["1", "2", "3"],
        "path": Path("/custom/path"),
    }
    expected = """
dict:
  str: mystr
  int: 123
list_of_int:
- 1
- 2
- 3
list_of_str:
- '1'
- '2'
- '3'
path: /custom/path
    """
    filepath = tmp_path / "test.yaml"

    test_module.dump_yaml(filepath, data)
    dumped_content = filepath.read_text(encoding="utf-8")
    assert dumped_content.strip() == expected.strip()


def test_load_yaml(tmp_path):
    data = """
dict:
  str: mystr
  int: 123
list_of_int:
- 1
- 2
- 3
list_of_str:
- '1'
- '2'
- '3'
path: /custom/path
    """
    expected = {
        "dict": {"str": "mystr", "int": 123},
        "list_of_int": [1, 2, 3],
        "list_of_str": ["1", "2", "3"],
        "path": "/custom/path",
    }
    filepath = tmp_path / "test.yaml"
    filepath.write_text(data, encoding="utf-8")

    loaded_data = test_module.load_yaml(filepath)
    assert loaded_data == expected


def test_dump_jaon_load_json_roundtrip(tmp_path):
    data = {
        "dict": {"str": "mystr", "int": 123},
        "list_of_int": [1, 2, 3],
        "list_of_str": ["1", "2", "3"],
        "path": "/custom/path",
    }
    filepath = tmp_path / "test.json"

    test_module.dump_json(filepath, data, indent=None)
    loaded_data = test_module.load_json(filepath)

    assert loaded_data == data


@pytest.mark.parametrize(
    "path, start, expected",
    [
        ("path/1", "path/2", "../1"),
        ("/path/1", "/path/2", "../1"),
        ("/path/to/1", "/path/2", "../to/1"),
        ("/path/1", "/path/to/2", "../../1"),
    ],
)
def test_relpath(path, start, expected):
    result = test_module.relpath(path, start=start)
    assert result == Path(expected)


@pytest.mark.parametrize(
    "d, expected",
    [
        (
            {},
            [],
        ),
        (
            {"key1": []},
            [],
        ),
        (
            {"key1": [100, 200]},
            [
                (("key1", 100, 0),),
                (("key1", 200, 1),),
            ],
        ),
        (
            {"key1": [100, 200], "key2": []},
            [],
        ),
        (
            {"key1": [100, 200], "key2": [300]},
            [
                (("key1", 100, 0), ("key2", 300, 0)),
                (("key1", 200, 1), ("key2", 300, 0)),
            ],
        ),
        (
            {"key1": [100, 200], "key2": [300, 400]},
            [
                (("key1", 100, 0), ("key2", 300, 0)),
                (("key1", 100, 0), ("key2", 400, 1)),
                (("key1", 200, 1), ("key2", 300, 0)),
                (("key1", 200, 1), ("key2", 400, 1)),
            ],
        ),
    ],
)
def test_dict_product(d, expected):
    result = test_module.dict_product(d)
    assert list(result) == expected


def test_extract_items():
    obj = {
        "latency": {"params": {"onset": False}},
        "decay": {"params": {"ratio": [0.25, 0.5, 0.75]}},
        "baseline_PSTH": {"params": {"bin_size": 0.5, "sigma": 0, "offset": -6}},
    }
    expected = [
        ("latency.params.onset", False),
        ("decay.params.ratio", [0.25, 0.5, 0.75]),
        ("baseline_PSTH.params.bin_size", 0.5),
        ("baseline_PSTH.params.sigma", 0),
        ("baseline_PSTH.params.offset", -6),
    ]
    result = list(test_module.extract_items(obj))
    assert result == expected


@pytest.mark.parametrize(
    "iterable, expected",
    [
        ([], True),
        ([1], True),
        ([1, 1], True),
        ([1, 1, 2], False),
        ("", True),
        ("aaaaaa", True),
        ("aaaaaba", False),
        ([[123, 456]], True),
        ([[123, 456], [123, 456]], True),
        ([[123, 456], [123, 456], [123, 999]], False),
    ],
)
def test_all_equal(iterable, expected):
    # with a list
    result = test_module.all_equal(iterable)
    assert result == expected
    # with an iterator
    result = test_module.all_equal(iter(iterable))
    assert result == expected


def test_copy_config(tmp_path):
    src = TEST_DATA_PATH / "analysis" / "analysis_config_01_relative.yaml"
    dst = tmp_path / "analysis_config.yaml"
    test_module.copy_config(src, dst)

    assert dst.is_file()
    src_config = yaml.safe_load(src.read_text())
    dst_config = yaml.safe_load(dst.read_text())
    src_sim_campaign_path = Path(src_config["simulation_campaign"])
    dst_sim_campaign_path = Path(dst_config["simulation_campaign"])
    assert src_sim_campaign_path.is_absolute() is False
    assert dst_sim_campaign_path.is_absolute() is True
    assert (src.parent / src_sim_campaign_path).resolve() == dst_sim_campaign_path.resolve()


def test_get_shmdir(monkeypatch, tmp_path):
    monkeypatch.setenv("SHMDIR", str(tmp_path))
    shmdir = test_module.get_shmdir()
    assert shmdir == tmp_path

    monkeypatch.delenv("SHMDIR")
    shmdir = test_module.get_shmdir()
    assert shmdir is None

    monkeypatch.setenv("SHMDIR", str(tmp_path / "non-existent"))
    with pytest.raises(RuntimeError, match="SHMDIR must be set to an existing directory"):
        test_module.get_shmdir()



from unittest.mock import MagicMock

import pandas as pd
import pytest

from tests.unit.utils import TEST_NODE_SETS_FILE


def _get_cells():
    """Return a DataFrame as returned by circuit.nodes[population].get()."""
    return pd.DataFrame(
        [
            {
                "layer": "1",
                "mtype": "L1_DAC",
                "etype": "cNAC",
                "region": "S1FL",
                "synapse_class": "INH",
                "x": 4497.1,
                "y": -1404.7,
                "z": -1710.8,
            },
            {
                "layer": "2",
                "mtype": "L2_TPC:A",
                "etype": "cADpyr",
                "region": "S1FL",
                "synapse_class": "EXC",
                "x": 4592.3,
                "y": -1351.1,
                "z": -1987.2,
            },
            {
                "layer": "4",
                "mtype": "L4_BP",
                "etype": "cNAC",
                "region": "S1FL",
                "synapse_class": "INH",
                "x": 3953.9,
                "y": -1279.3,
                "z": -2143.9,
            },
        ],
        index=pd.Index([100, 200, 300], name="node_ids"),
    )


@pytest.fixture
def mock_circuit():
    """Simplified mock of circuit, providing only get() and ids() for a node population."""
    mock = MagicMock()
    mock.node_sets_file = str(TEST_NODE_SETS_FILE)
    df = _get_cells()
    # circuit.nodes[population]
    mock_population = mock.nodes.__getitem__.return_value
    # circuit.nodes[population].get()
    mock_population.get.return_value = df
    # circuit.nodes[population].ids()
    mock_population.ids.return_value = df.index.to_numpy()
    return mock


@pytest.fixture
def mock_simulation():
    """Simplified mock of simulation, providing only get() for a spikes population."""
    mock = MagicMock()
    # simulation.spikes[population]
    mock_population = mock.spikes.__getitem__.return_value
    # simulation.spikes[population].get() returns a pandas.Series
    # with the spiking node_ids indexed by sorted spike time
    mock_population.get.return_value = pd.Series(
        [200, 100, 200, 200],
        index=[100.1, 120.2, 120.2, 150.0],
    )
    return mock



from unittest.mock import Mock, PropertyMock

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal, assert_series_equal

from blueetl.config.analysis_model import NeuronClassConfig
from blueetl.constants import CIRCUIT, CIRCUIT_ID, GID, NEURON_CLASS, SIMULATION, SIMULATION_ID
from blueetl.extract.neurons import Neurons
from blueetl.utils import ensure_dtypes
from tests.unit.utils import TEST_NODE_SETS_FILE_EXTRA


def test_neurons_from_simulations(mock_circuit):
    # set the seed because the neurons may be chosen randomly
    np.random.seed(0)
    mock_simulations_df = PropertyMock(
        return_value=pd.DataFrame(
            [
                {SIMULATION_ID: 0, CIRCUIT_ID: 0, SIMULATION: Mock(), CIRCUIT: mock_circuit},
                {SIMULATION_ID: 1, CIRCUIT_ID: 0, SIMULATION: Mock(), CIRCUIT: mock_circuit},
            ]
        )
    )
    mock_simulations = Mock()
    type(mock_simulations).df = mock_simulations_df
    neuron_classes = {
        "L1_INH": NeuronClassConfig.model_validate(
            {
                "population": "thalamus_neurons",
                "query": {"layer": ["1"], "synapse_class": ["INH"]},
            }
        ),
        "MY_GIDS": NeuronClassConfig(**{"population": "thalamus_neurons", "node_id": [200, 300]}),
        "EMPTY": NeuronClassConfig.model_validate(
            {"population": "thalamus_neurons", "query": {"layer": ["999"]}}
        ),
        "LIMITED": NeuronClassConfig.model_validate(
            {
                "population": "thalamus_neurons",
                "query": {"synapse_class": ["INH"]},
                "limit": 1,
                "node_set": "ExtraLayer2",
                "node_sets_file": TEST_NODE_SETS_FILE_EXTRA,
            }
        ),
    }
    result = Neurons.from_simulations(
        simulations=mock_simulations,
        neuron_classes=neuron_classes,
    )
    expected_df = pd.DataFrame(
        [
            {
                "circuit_id": 0,
                "neuron_class": "L1_INH",
                "gid": 100,
                "neuron_class_index": 0,
            },
            {
                "circuit_id": 0,
                "neuron_class": "MY_GIDS",
                "gid": 200,
                "neuron_class_index": 0,
            },
            {
                "circuit_id": 0,
                "neuron_class": "MY_GIDS",
                "gid": 300,
                "neuron_class_index": 1,
            },
            {
                "circuit_id": 0,
                "neuron_class": "LIMITED",
                "gid": 300,
                "neuron_class_index": 0,
            },
        ],
    )
    expected_df = expected_df.sort_values([CIRCUIT_ID, NEURON_CLASS, GID], ignore_index=True)
    expected_df = ensure_dtypes(expected_df)
    assert isinstance(result, Neurons)
    assert_frame_equal(result.df, expected_df)
    assert mock_circuit.nodes.__getitem__.return_value.get.call_count == 2
    assert mock_simulations_df.call_count == 1


def test_neurons_from_simulations_without_neurons(mock_circuit):
    mock_simulations_df = PropertyMock(
        return_value=pd.DataFrame(
            [
                {SIMULATION_ID: 0, CIRCUIT_ID: 0, SIMULATION: Mock(), CIRCUIT: mock_circuit},
                {SIMULATION_ID: 1, CIRCUIT_ID: 0, SIMULATION: Mock(), CIRCUIT: mock_circuit},
            ]
        )
    )
    mock_simulations = Mock()
    type(mock_simulations).df = mock_simulations_df
    neuron_classes = {
        "EMPTY": NeuronClassConfig(
            **{"population": "thalamus_neurons", "query": {"layer": ["999"]}}
        ),
    }

    with pytest.raises(RuntimeError, match="No data extracted to Neurons"):
        Neurons.from_simulations(simulations=mock_simulations, neuron_classes=neuron_classes)
    assert mock_circuit.nodes.__getitem__.return_value.get.call_count == 1
    assert mock_simulations_df.call_count == 1


def test_neurons_count_by_neuron_class():
    df = pd.DataFrame(
        [
            {
                "circuit_id": 0,
                "neuron_class": "NC1",
                "gid": 100,
                "neuron_class_index": 0,
            },
            {
                "circuit_id": 0,
                "neuron_class": "NC1",
                "gid": 200,
                "neuron_class_index": 1,
            },
            {
                "circuit_id": 0,
                "neuron_class": "NC2",
                "gid": 100,
                "neuron_class_index": 0,
            },
            {
                "circuit_id": 1,
                "neuron_class": "NC1",
                "gid": 100,
                "neuron_class_index": 0,
            },
        ],
    )
    neurons = Neurons.from_pandas(df)
    result = neurons.count_by_neuron_class()
    expected_series = ensure_dtypes(
        pd.DataFrame(
            [
                [0, "NC1", 2],
                [0, "NC2", 1],
                [1, "NC1", 1],
            ],
            columns=[CIRCUIT_ID, NEURON_CLASS, GID],
        )
    ).set_index([CIRCUIT_ID, NEURON_CLASS])[GID]

    assert_series_equal(result, expected_series)



import pytest

from blueetl.campaign import utils as test_module


@pytest.mark.parametrize(
    "include_empty, expected",
    [
        (
            True,
            [
                ({"ca": 1.0, "depolarization": 3.0}, None),
                ({"ca": 1.0, "depolarization": 4.0}, "/tmp/simple/uuid/1/simulation_config.json"),
                ({"ca": 2.0, "depolarization": 3.0}, "/tmp/simple/uuid/2/simulation_config.json"),
                ({"ca": 2.0, "depolarization": 4.0}, "/tmp/simple/uuid/3/simulation_config.json"),
            ],
        ),
        (
            False,
            [
                ({"ca": 1.0, "depolarization": 4.0}, "/tmp/simple/uuid/1/simulation_config.json"),
                ({"ca": 2.0, "depolarization": 3.0}, "/tmp/simple/uuid/2/simulation_config.json"),
                ({"ca": 2.0, "depolarization": 4.0}, "/tmp/simple/uuid/3/simulation_config.json"),
            ],
        ),
    ],
)
def test_campaign_sims(xarray_config_dict_simple, include_empty, expected):
    result = list(test_module.campaign_sims(xarray_config_dict_simple, include_empty=include_empty))
    assert result == expected


@pytest.mark.parametrize(
    "include_empty, expected",
    [
        (
            True,
            [
                ({"ca": 1.0, "depolarization": 3.0}, None),
                ({"ca": 2.0, "depolarization": 4.0}, "/tmp/coupled/uuid/1/simulation_config.json"),
            ],
        ),
        (
            False,
            [
                ({"ca": 2.0, "depolarization": 4.0}, "/tmp/coupled/uuid/1/simulation_config.json"),
            ],
        ),
    ],
)
def test_campaign_sims_coupled(xarray_config_dict_coupled, include_empty, expected):
    result = list(
        test_module.campaign_sims(xarray_config_dict_coupled, include_empty=include_empty)
    )
    assert result == expected


def test_campaign_sims_with_indices(xarray_config_dict_simple):
    result = list(test_module._campaign_sims_with_indices(xarray_config_dict_simple))
    expected = [
        (
            {"ca": 1.0, "depolarization": 3.0},
            {
                "ca_idx": 0,
                "ca_total": 2,
                "depolarization_idx": 0,
                "depolarization_total": 2,
                "idx": 0,
            },
            "",
        ),
        (
            {"ca": 1.0, "depolarization": 4.0},
            {
                "ca_idx": 0,
                "ca_total": 2,
                "depolarization_idx": 1,
                "depolarization_total": 2,
                "idx": 1,
            },
            "/tmp/simple/uuid/1/simulation_config.json",
        ),
        (
            {"ca": 2.0, "depolarization": 3.0},
            {
                "ca_idx": 1,
                "ca_total": 2,
                "depolarization_idx": 0,
                "depolarization_total": 2,
                "idx": 2,
            },
            "/tmp/simple/uuid/2/simulation_config.json",
        ),
        (
            {"ca": 2.0, "depolarization": 4.0},
            {
                "ca_idx": 1,
                "ca_total": 2,
                "depolarization_idx": 1,
                "depolarization_total": 2,
                "idx": 3,
            },
            "/tmp/simple/uuid/3/simulation_config.json",
        ),
    ]
    assert result == expected


def test_campaign_sims_with_indices_coupled(xarray_config_dict_coupled):
    with pytest.raises(ValueError, match="Please provide non-coupled coords sim campaign"):
        list(test_module._campaign_sims_with_indices(xarray_config_dict_coupled))


def test_campaign_sim_indices(xarray_config_dict_simple):
    result = test_module._campaign_sim_indices(xarray_config_dict_simple)
    assert result == [1, 2, 3]


def test_campaign_sim_indices_coupled(xarray_config_dict_coupled):
    result = test_module._campaign_sim_indices(xarray_config_dict_coupled)
    assert result == [1]


def test_campaign_sim_index_to_coords(xarray_config_dict_simple):
    result = test_module._campaign_sim_index_to_coords(xarray_config_dict_simple)
    assert result == {
        1: {"ca": 1.0, "depolarization": 4.0},
        2: {"ca": 2.0, "depolarization": 3.0},
        3: {"ca": 2.0, "depolarization": 4.0},
    }


def test_campaign_sim_index_to_coords_coupled(xarray_config_dict_coupled):
    result = test_module._campaign_sim_index_to_coords(xarray_config_dict_coupled)
    assert result == {
        1: {"ca": 2.0, "depolarization": 4.0},
    }



"""Common utilities."""

import hashlib
import importlib
import itertools
import json
import logging
import os.path
import time
from collections.abc import Iterable, Iterator
from contextlib import contextmanager
from functools import cache, cached_property
from pathlib import Path
from typing import Any, Callable, Optional, Union

import pandas as pd
import yaml

from blueetl.constants import DTYPES
from blueetl.types import StrOrPath

L = logging.getLogger(__name__)


class CachedPropertyMixIn:
    """MixIn to be used with classes using cached_property to be skipped when pickled."""

    def __getstate__(self) -> dict:
        """Get the object state when the object is pickled."""
        return {
            key: value
            for key, value in self.__dict__.items()
            if not isinstance(getattr(self.__class__, key, None), cached_property)
        }

    def __setstate__(self, state: dict) -> None:
        """Set the object state when the object is unpickled."""
        self.__dict__.update(state)


@contextmanager
def timed(log: Callable[[str], None], msg: str) -> Iterator[list[str]]:
    """Context manager to log the execution time using the specified logger function.

    The yielded list of messages can be used to replace the existing message,
    or inject additional strings to the final log.
    """
    log(f"{msg}...")
    start_time = time.monotonic()
    status = "FAILED"
    messages = [msg]
    try:
        yield messages
        status = "DONE"
    finally:
        elapsed = time.monotonic() - start_time
        log(f"{' '.join(messages)} [{status} in {elapsed:.2f} seconds]")


def setup_logging(loglevel: Union[int, str], logformat: Optional[str] = None, **logparams) -> None:
    """Setup logging."""
    logformat = logformat or "%(asctime)s %(levelname)s %(name)s: %(message)s"
    logging.basicConfig(format=logformat, level=loglevel, **logparams)


def load_yaml(filepath: StrOrPath) -> Any:
    """Load from YAML file."""
    with open(filepath, encoding="utf-8") as f:
        # Any conversion when loading back the values can be done later,
        # so that the loaded object can be validated using jsonschema.
        return yaml.load(f, Loader=yaml.SafeLoader)


def dump_yaml(filepath: StrOrPath, data: Any, **kwargs) -> None:
    """Dump to YAML file."""
    with open(filepath, "w", encoding="utf-8") as f:
        # The custom dumper dumps some unsupported types (for example Path) as simple strings.
        yaml.dump(data, stream=f, sort_keys=False, Dumper=_get_internal_yaml_dumper(), **kwargs)


def load_json(filepath: StrOrPath, *, encoding: str = "utf-8", **kwargs) -> Any:
    """Load from JSON file."""
    with open(filepath, encoding=encoding) as f:
        return json.load(f, **kwargs)


def dump_json(
    filepath: StrOrPath, data: Any, *, encoding: str = "utf-8", indent: int = 2, **kwargs
) -> None:
    """Dump to JSON file."""
    with open(filepath, mode="w", encoding=encoding) as fp:
        json.dump(data, fp, indent=indent, **kwargs)


def relpath(path: StrOrPath, start: StrOrPath) -> Path:
    """Return a relative filepath to path from the start directory.

    In Python>=3.12 it would be possible to use ``Path.relative_to`` with walk_up=True.
    """
    return Path(os.path.relpath(path, start=start))


def ensure_list(x: Any) -> list:
    """Always return a list from the given argument."""
    if isinstance(x, list):
        return x
    if isinstance(x, tuple):
        return list(x)
    return [x]


def ensure_dtypes(
    df: pd.DataFrame, desired_dtypes: Optional[dict[str, Any]] = None
) -> pd.DataFrame:
    """Return a DataFrame with the columns and index cast to the desired types.

    Args:
        df: original Pandas DataFrame.
        desired_dtypes: dict of names and desired dtypes. If None, the predefined dtypes are used.
            If the dict contains names not present in the columns or in the index, they are ignored.

    Returns:
        The same DataFrame is modified and returned, if it's possible to avoid making a copy.
    """
    if desired_dtypes is None:
        desired_dtypes = DTYPES
    # convert the columns data types
    if dtypes := {
        k: desired_dtypes[k]
        for k in df.columns
        if k in desired_dtypes and desired_dtypes[k] != df.dtypes.at[k]
    }:
        df = df.astype(dtypes)
    # convert the index data types
    if dtypes := {
        k: desired_dtypes[k]
        for k in df.index.names
        if k in desired_dtypes and desired_dtypes[k] != df.index.etl.dtypes.at[k]
    }:
        df.index = df.index.etl.astype(dtypes)
    return df


def import_by_string(full_name: str) -> Callable:
    """Import and return a function by name.

    Args:
        full_name: full name of the function, using dot as a separator if in a submodule.

    Returns:
        The imported function.
    """
    module_name, _, func_name = full_name.rpartition(".")
    return getattr(importlib.import_module(module_name), func_name)


def resolve_path(*paths: StrOrPath, symlinks: bool = False) -> Path:
    """Make the path absolute and return a new path object.

    It may be different from calling Path.resolve(), because Path.resolve() always resolve symlinks.
    It may be different from calling Path.absolute(), because Path.absolute() doesn't remove the
    relative paths. For example, Path('/tmp/..').absolute() == PosixPath('/tmp/..').
    """
    if symlinks:
        # resolve any symlinks
        return Path(*paths).resolve()
    # does not resolve symbolic links
    return Path(os.path.abspath(Path(*paths)))


def checksum(filepath: StrOrPath, chunk_size: int = 65536) -> str:
    """Calculate and return the checksum of the given file."""
    filehash = hashlib.blake2b()
    with open(filepath, "rb") as f:
        while chunk := f.read(chunk_size):
            filehash.update(chunk)
    return filehash.hexdigest()


def checksum_str(s: str) -> str:
    """Calculate and return the checksum of the given string."""
    return hashlib.blake2b(s.encode("utf-8")).hexdigest()


def checksum_json(obj: Any) -> str:
    """Calculate and return the checksum of the given object converted to json."""
    return checksum_str(json.dumps(obj, sort_keys=True))


@cache
def _get_internal_yaml_dumper() -> type[yaml.SafeDumper]:
    """Return the custom internal yaml dumper class."""
    # pylint: disable=import-outside-toplevel
    # imported here because optional
    from pydantic import BaseModel

    _representers = {
        Path: str,
        BaseModel: lambda data: data.dict(),
    }

    class Dumper(yaml.SafeDumper):
        """Custom YAML Dumper."""

        def represent_data(self, data):
            for cls in type(data).__mro__:
                if func := _representers.get(cls):
                    return self.represent_data(func(data))
            return super().represent_data(data)

    return Dumper


def dict_product(d: dict) -> Iterator[tuple]:
    """Iterate over the product of the values of the given dict.

    Yield tuples of tuples, where each item is composed by:

    - key
    - value
    - index of the key in the original list

    Example:
        Passing the dictionary:

        .. code-block:: python

            d = {"a": ["a1", "a2"], "b": ["b1", "b2"]}

        Yields tuples from:

        .. code-block:: python

            [
                (("a", "a1", 0), ("b", "b1", 0)),
                (("a", "a1", 0), ("b", "b2", 1)),
                (("a", "a2", 1), ("b", "b1", 0)),
                (("a", "a2", 1), ("b", "b2", 1)),
            ]

    """
    if d:
        yield from itertools.product(
            *[[(key, v, i) for i, v in enumerate(values)] for key, values in d.items()]
        )


def extract_items(obj: dict[str, Any], path: Optional[str] = None) -> Iterator[tuple[str, Any]]:
    """Yield tuples (path, item) from the traversal of the given dict.

    Each yielded path is obtained from the concatenation of the nested keys, separated by ".".

    All the keys of the dicts are expected to be strings, not containing ".".

    For examples, iterating over the result of:

    .. code-block:: python

        extract_items(
            {
                "latency": {"params": {"onset": False}},
                "decay": {"params": {"ratio": [0.25, 0.5, 0.75]}},
                "baseline_PSTH": {"params": {"bin_size": 0.5, "sigma": 0, "offset": -6}},
            }
        )

    would yield the following tuples:

    .. code-block:: python

        ("latency.params.onset", False)
        ("decay.params.ratio",  [0.25, 0.5, 0.75])
        ("baseline_PSTH.params.bin_size", 0.5)
        ("baseline_PSTH.params.sigma", 0)
        ("baseline_PSTH.params.offset", -6)

    """
    if isinstance(obj, dict):
        for k, v in obj.items():
            assert isinstance(k, str)
            assert "." not in k
            yield from extract_items(v, path=k if path is None else f"{path}.{k}")
    else:
        yield path, obj


def all_equal(iterable: Iterable) -> bool:
    """Return True if all the items of the given iterable are equal, False otherwise.

    Notes:
    - The items doesn't need to be hashable, because they are compared with equality.
    - The function returns True also when the iterable doesn't contain any item.
    """
    prev = None
    for n, item in enumerate(iterable):
        if n > 0 and item != prev:
            return False
        prev = item
    return True


def import_optional_dependency(name: str) -> Any:
    """Import an optional dependency.

    If a dependency is missing, an ImportError with a custom message is raised. Based on:
    https://github.com/pandas-dev/pandas/blob/0d853e77/pandas/compat/_optional.py#L85

    Args:
        name: The module name.

    Returns:
        ModuleType: The imported module, if found.
    """
    try:
        module = importlib.import_module(name)
    except ImportError as ex:
        msg = f"Missing optional dependency {name!r}. Use pip to install it."
        raise ImportError(msg) from ex

    return module


def copy_config(src: StrOrPath, dst: StrOrPath) -> None:
    """Copy the analysis configuration file to a different location.

    If the simulation_campaign path is relative, then it's resolved
    relatively to the directory containing the original configuration file.

    The output path, instead, is not resolved even when it's relative.

    Note that any comment present in the original file is not preserved.
    """
    src = Path(src)
    config = load_yaml(src)
    config["simulation_campaign"] = resolve_path(src.parent, config["simulation_campaign"])
    dump_yaml(dst, config, default_flow_style=None)


def get_shmdir() -> Optional[Path]:
    """Return the shared memory directory, or None if not set."""
    shmdir = os.getenv("SHMDIR")
    if not shmdir:
        L.warning("SHMDIR should be set to the shared memory directory")
        return None
    shmdir = Path(shmdir)
    if not shmdir.is_dir():
        raise RuntimeError("SHMDIR must be set to an existing directory")
    return shmdir



"""Cache Manager."""

import errno
import fcntl
import logging
import os
import shutil
from abc import abstractmethod
from copy import deepcopy
from dataclasses import dataclass
from functools import wraps
from typing import Generic, Optional, Protocol, TypeVar

import pandas as pd
from blueetl_core.utils import is_subfilter

from blueetl.campaign.config import SimulationCampaign
from blueetl.config.analysis_model import CacheConfig, FeaturesConfig, SingleAnalysisConfig
from blueetl.store.base import BaseStore
from blueetl.store.feather import FeatherStore
from blueetl.store.parquet import ParquetStore
from blueetl.utils import dump_yaml, load_yaml

L = logging.getLogger(__name__)

ConfigT = TypeVar("ConfigT", SingleAnalysisConfig, SimulationCampaign)


@dataclass
class CoupledCache(Generic[ConfigT]):
    """Container of cached and actual configurations."""

    cached: Optional[ConfigT]
    actual: ConfigT


def _raise_if(**attrs):
    """Raise if the decorated method is called and all the attrs are equal to the given values."""

    def decorator(f):
        @wraps(f)
        def wrapper(self, *args, **kwargs):
            if all(getattr(self, attr) == value for attr, value in attrs.items()):
                raise CacheError(
                    f"Method {self.__class__.__name__}.{f.__name__} "
                    f"cannot be called when the attributes are: {attrs}"
                )
            return f(self, *args, **kwargs)

        return wrapper

    return decorator


class CacheError(Exception):
    """Cache error raised when a read-only cache is written."""


class LockManagerProtocol(Protocol):
    """Lock Manager Interface."""

    @property
    @abstractmethod
    def locked(self) -> bool:
        """Return the lock status."""

    @abstractmethod
    def lock(self, mode: int) -> None:
        """Lock."""

    @abstractmethod
    def unlock(self) -> None:
        """Unlock."""


class LockManager(LockManagerProtocol):
    """Lock Manager.

    On Linux, the flock call is handled locally, and the underlying filesystem (GPFS) does not get
    any notification that locks are being set. Therefore, GPFS cannot enforce locks across nodes.
    """

    LOCK_EX = fcntl.LOCK_EX  # exclusive lock
    LOCK_SH = fcntl.LOCK_SH  # shared lock

    def __init__(self, path: os.PathLike) -> None:
        """Initialize the object.

        Args:
            path: path to an existing directory to be used for locking.
        """
        self._path = path
        self._fd: Optional[int] = None

    @property
    def locked(self) -> bool:
        """Return True if the lock manager is locking the cache, False otherwise."""
        return self._fd is not None

    def lock(self, mode: int) -> None:
        """Lock the directory.

        Args:
            mode: LockManager.LOCK_EX for exclusive lock, or LockManager.LOCK_SH for shared lock.
        """
        if self._fd is None:
            self._fd = os.open(self._path, os.O_RDONLY)
        try:
            fcntl.flock(self._fd, mode | fcntl.LOCK_NB)
        except OSError as ex:
            os.close(self._fd)
            self._fd = None
            if ex.errno in (errno.EACCES, errno.EAGAIN):
                # the lock cannot be acquired
                raise CacheError(f"Another process is locking {self._path}") from None
            # re-raise any other type of OSError
            raise

    def unlock(self) -> None:
        """Unlock the directory."""
        if self._fd is not None:
            os.close(self._fd)
            self._fd = None


class DummyLockManager(LockManagerProtocol):
    """Dummy Lock Manager."""

    @property
    def locked(self) -> bool:
        """Always return True."""
        return True

    def lock(self, mode: int) -> None:
        """Pretend to lock."""

    def unlock(self) -> None:
        """Pretend to unlock."""


class CacheManager:
    """Cache Manager."""

    def __init__(
        self,
        cache_config: CacheConfig,
        analysis_config: SingleAnalysisConfig,
        simulations_config: SimulationCampaign,
    ) -> None:
        """Initialize the object.

        Args:
            cache_config: cache configuration dict.
            analysis_config: analysis configuration dict.
            simulations_config: simulations campaign configuration.
        """
        self._output_dir = cache_config.path
        if cache_config.clear:
            self._clear_cache()
        repo_dir = self._output_dir / "repo"
        features_dir = self._output_dir / "features"
        config_dir = self._output_dir / "config"
        for new_dir in repo_dir, features_dir, config_dir:
            new_dir.mkdir(exist_ok=True, parents=True)

        store_classes: dict[str, type[BaseStore]] = {
            "parquet": ParquetStore,
            "feather": FeatherStore,
        }
        store_class = store_classes[cache_config.store_type]

        self._readonly = cache_config.readonly
        self._skip_features = cache_config.skip_features

        self._version = 1
        self._repo_store = store_class(repo_dir)
        self._features_store = store_class(features_dir)

        self._lock_manager: LockManagerProtocol = LockManager(self._output_dir)
        self._lock_manager.lock(mode=LockManager.LOCK_SH if self._readonly else LockManager.LOCK_EX)

        self._cached_analysis_config_path = config_dir / "analysis_config.cached.yaml"
        self._cached_simulations_config_path = config_dir / "simulations_config.cached.yaml"
        self._cached_checksums_path = config_dir / "checksums.cached.yaml"

        self._analysis_configs = CoupledCache[SingleAnalysisConfig](
            cached=self._load_cached_analysis_config(),
            actual=analysis_config,
        )
        self._simulations_configs = CoupledCache[SimulationCampaign](
            cached=self._load_cached_simulations_config(),
            actual=simulations_config,
        )
        self._cached_checksums = self._load_cached_checksums()
        self._initialize_cache()

    def _clear_cache(self):
        """Remove the cache directory if it exists."""
        L.info("Removing cache if it exists: %s", self._output_dir)
        shutil.rmtree(self._output_dir, ignore_errors=True)

    def __getstate__(self) -> dict:
        """Get the object state when the object is pickled."""
        return self.__dict__ | {"_lock_manager": None}

    def __setstate__(self, state: dict) -> None:
        """Set the object state when the object is unpickled."""
        self.__dict__.update(state)
        # The unpickled object must always be readonly, even when the pickled object isn't.
        self._readonly = True
        # A new lock isn't created in the subprocess b/c we want to be able to read the cache.
        self._lock_manager = DummyLockManager()

    @property
    def locked(self) -> bool:
        """Return True if the cache manager is locking the cache, False otherwise."""
        return self._lock_manager.locked

    def close(self) -> None:
        """Close the cache manager and unlock the lock directory.

        After calling this method, the Cache Manager instance shouldn't be used anymore.
        """
        self._lock_manager.unlock()

    @property
    def readonly(self) -> bool:
        """Return True if the cache manager is set to read-only, False otherwise."""
        return self._readonly

    @_raise_if(locked=False)
    def to_readonly(self) -> "CacheManager":
        """Return a read-only copy of the object.

        The returned object will raise an exception if any writing method is called.
        """
        obj = deepcopy(self)
        # readonly is always True because of __getstate__ and __setstate__
        assert obj.readonly is True
        return obj

    def _load_cached_analysis_config(self) -> Optional[SingleAnalysisConfig]:
        """Load the cached analysis config if it exists."""
        path = self._cached_analysis_config_path
        return SingleAnalysisConfig.load(path) if path.exists() else None

    def _load_cached_simulations_config(self) -> Optional[SimulationCampaign]:
        """Load the cached simulations config if it exists."""
        path = self._cached_simulations_config_path
        return SimulationCampaign.load(path) if path.exists() else None

    def _load_cached_checksums(self) -> dict:
        """Load the cached checksums, or return null checksums if the file doesn't exist."""
        path = self._cached_checksums_path
        checksums = None
        if path.exists():
            L.info("Loading cached checksums from %s", path)
            checksums = load_yaml(path)
            if checksums["version"] != self._version:
                L.warning(
                    "Incompatible cache version %s != %s, the cache is being deleted",
                    checksums["version"],
                    self._version,
                )
                checksums = None
        if not checksums:
            checksums = {
                "version": self._version,
                "repo": {
                    "simulations": None,
                    "neurons": None,
                    "neuron_classes": None,
                    "windows": None,
                    "report": None,
                },
                "features": {},
            }
        return checksums

    @_raise_if(readonly=True)
    def _dump_analysis_config(self) -> None:
        """Write the cached analysis config to file."""
        path = self._cached_analysis_config_path
        self._analysis_configs.actual.dump(path)

    @_raise_if(readonly=True)
    def _dump_simulations_config(self) -> None:
        """Write the cached simulations config to file."""
        path = self._cached_simulations_config_path
        self._simulations_configs.actual.dump(path)

    @_raise_if(readonly=True)
    def _dump_cached_checksums(self) -> None:
        """Write the cached checksums to file."""
        path = self._cached_checksums_path
        dump_yaml(path, self._cached_checksums)

    def _invalidate_cached_checksums(self, names: Optional[set[str]] = None) -> None:
        """Invalidate the checksums for the given names.

        This function takes into consideration the dependencies between the dataframes,
        so the cache of all the dataframes depending on the given dataframes are invalidated,
        without the need to specify all of them in ``names``.

        The special name ``features`` can be specified to invalidate all the features.

        If ``names`` is None or empty, all the checksums are invalidated.

        Args:
            names: set of names to be invalidated.
        """

        def _invalidate_repo(_name):
            self._cached_checksums["repo"][_name] = None

        def _invalidate_features(config_checksum):
            d = self._cached_checksums["features"][config_checksum]
            for _name in d:
                d[_name] = None

        def _invalidate_all_features():
            for config_checksum in self._cached_checksums["features"]:
                _invalidate_features(config_checksum)

        ordered_names = [
            "simulations",
            "neurons",
            "neuron_classes",
            "windows",
            "report",
            "features",
        ]
        assert not names or names.issubset(ordered_names), "Invalid names specified."
        invalidated: list[str] = []
        for name in ordered_names:
            if invalidated or not names or name in names:
                if name == "features":
                    _invalidate_all_features()
                else:
                    _invalidate_repo(name)
                invalidated.append(name)
        L.info("Invalid cache: %s", invalidated)

    def _check_config_cache(self) -> bool:
        """Compare the cached and actual configurations.

        Only the cache of incompatible dataframes is invalidated.

        Returns:
            True if the cache is valid and complete, False otherwise.
        """
        if not self._analysis_configs.cached or not self._simulations_configs.cached:
            L.debug("The cached configurations have not been found")
            self._invalidate_cached_checksums()
            return False

        # check the criteria used to filter the simulations
        if not self._is_subfilter(strict=False):
            L.debug("The simulations filter is less specific, so the cache cannot be used")
            self._invalidate_cached_checksums()
            return False

        # check the simulations config
        if self._simulations_configs.cached != self._simulations_configs.actual:
            L.debug("The cached simulation campaign configurations is invalid")
            self._invalidate_cached_checksums({"simulations"})
            return False

        # check the extraction config for changed keys, and invalidate the affected names
        keys_and_affected_names = [
            ({"neuron_classes"}, {"neurons", "neuron_classes"}),
            ({"windows", "trial_steps"}, {"windows", "report"}),
            ({"report"}, {"report"}),
        ]
        for keys, names in keys_and_affected_names:
            if any(
                getattr(self._analysis_configs.cached.extraction, k)
                != getattr(self._analysis_configs.actual.extraction, k)
                for k in keys
            ):
                L.debug("The following cached objects are invalid: %s", names)
                self._invalidate_cached_checksums(names)
                return False

        # check the features config
        valid_checksums = set()
        for features_config in self._analysis_configs.actual.features:
            config_checksum = features_config.checksum()
            if config_checksum in self._cached_checksums["features"] and all(
                self._cached_checksums["features"][config_checksum].values()
            ):
                valid_checksums.add(config_checksum)

        # invalidate the invalid features checksums
        is_valid = True
        for config_checksum, d in self._cached_checksums["features"].items():
            if config_checksum not in valid_checksums:
                is_valid = False
                for _name in d:
                    d[_name] = None

        return is_valid

    def _check_cached_repo_files(self) -> set[str]:
        """Determine the cached repo files to be deleted b/c the checksum is None or different.

        Returns:
            set of repository names to be deleted.
        """
        to_be_deleted = set()
        for name, file_checksum in self._cached_checksums["repo"].items():
            if not file_checksum or file_checksum != self._repo_store.checksum(name):
                to_be_deleted.add(name)
        return to_be_deleted

    @_raise_if(readonly=True)
    def _delete_cached_repo_files(self, to_be_deleted: set[str]) -> None:
        """Delete the given repository files.

        Args:
            to_be_deleted: set of repository names to be deleted.
        """
        for name in to_be_deleted:
            L.info("Deleting cached repo %s", name)
            self._repo_store.delete(name)
            del self._cached_checksums["repo"][name]

    def _check_cached_features_files(self) -> set[str]:
        """Determine the cached features files to be deleted b/c the checksum is None or different.

        Returns:
            set of features checksums to be deleted.
        """
        to_be_deleted = set()
        for config_checksum, checksums_by_name in self._cached_checksums["features"].items():
            for name, file_checksum in checksums_by_name.items():
                if not file_checksum or file_checksum != self._features_store.checksum(name):
                    to_be_deleted.add(config_checksum)
                    break
        return to_be_deleted

    @_raise_if(readonly=True)
    def _delete_cached_features_files(self, to_be_deleted) -> None:
        """Delete the files corresponding to the given features checksums.

        Args:
            to_be_deleted: set of features checksums to be deleted.
        """
        for config_checksum in to_be_deleted:
            # delete every feature generated with the same configuration
            for name in self._cached_checksums["features"].pop(config_checksum):
                L.info("Deleting invalid cached features %s/%s", config_checksum[:8], name)
                self._features_store.delete(name)

    def _initialize_cache(self) -> None:
        """Initialize the cache."""
        L.info("Initialize cache")
        is_config_cache_valid = self._check_config_cache()
        repo_to_be_deleted = self._check_cached_repo_files()
        features_to_be_deleted = self._check_cached_features_files()
        if repo_to_be_deleted:
            self._delete_cached_repo_files(repo_to_be_deleted)
        if features_to_be_deleted:
            self._delete_cached_features_files(features_to_be_deleted)
        if not is_config_cache_valid:
            self._dump_analysis_config()
            self._dump_simulations_config()
            self._dump_cached_checksums()

    @_raise_if(locked=False)
    def is_repo_cached(self, name: str) -> bool:
        """Return whether a specific repo dataframe is present in the cache."""
        # the checksums have been checked in _initialize_cache/_delete_cached_repo_files,
        # so they are not calculate again here
        return bool(
            self._cached_checksums["repo"].get(name) and self._repo_store.path(name).is_file()
        )

    @_raise_if(locked=False)
    def load_repo(self, name: str) -> Optional[pd.DataFrame]:
        """Load a specific repo dataframe from the cache.

        Args:
            name: name of the repo dataframe.

        Returns:
            The loaded dataframe, or None if it's not cached.
        """
        is_cached = self.is_repo_cached(name)
        return self._repo_store.load(name) if is_cached else None

    @_raise_if(readonly=True)
    @_raise_if(locked=False)
    def dump_repo(self, df: pd.DataFrame, name: str) -> None:
        """Write a specific repo dataframe to the cache.

        Args:
            df: dataframe to be saved.
            name: name of the repo dataframe.
        """
        self._repo_store.dump(df, name)
        self._cached_checksums["repo"][name] = self._repo_store.checksum(name)
        self._dump_cached_checksums()

    @_raise_if(locked=False)
    def get_cached_features_checksums(
        self, features_config: FeaturesConfig
    ) -> dict[str, dict[str, str]]:
        """Return the cached features checksums, or an empty dict if the cache doesn't exist."""
        config_checksum = features_config.checksum()
        cached = self._cached_checksums["features"].get(config_checksum, {})
        return cached

    @_raise_if(locked=False)
    def load_features(self, features_config: FeaturesConfig) -> Optional[dict[str, pd.DataFrame]]:
        """Load features dataframes from the cache.

        The cache key is determined by the hash of features_config.

        Args:
            features_config: configuration dict of the features to be loaded.

        Returns:
            Dict of dataframes, or None if they are not cached.
        """
        cached_checksums = self.get_cached_features_checksums(features_config)
        features = {}
        # the checksums have been checked in _initialize_cache/_delete_cached_features_files,
        # so they are not calculate again here
        for name, file_checksum in cached_checksums.items():
            assert file_checksum is not None
            features[name] = self._features_store.load(name)
            assert features[name] is not None
        return features or None

    @_raise_if(readonly=True)
    @_raise_if(locked=False)
    def dump_features(
        self, features_dict: dict[str, pd.DataFrame], features_config: FeaturesConfig
    ) -> None:
        """Write features dataframes to the cache.

        The cache key is determined by the hash of features_config.

        Args:
            features_dict: dict of features to be written.
            features_config: configuration dict of the features to be written.
        """
        if self._skip_features:
            L.info("Skipping writing features to cache")
            return
        config_checksum = features_config.checksum()
        old_checksums = self._cached_checksums["features"].pop(config_checksum, None)
        new_checksums = self._cached_checksums["features"][config_checksum] = {}
        for name, feature in features_dict.items():
            self._features_store.dump(feature, name)
            new_checksums[name] = self._features_store.checksum(name)
        if old_checksums is not None:
            assert (
                len(set(old_checksums).difference(new_checksums)) == 0
            ), "Some features have been found only in the old cached data"
            assert (
                len(set(new_checksums).difference(old_checksums)) == 0
            ), "Some features have been found only in the new cached data"
        self._dump_cached_checksums()

    def _is_subfilter(self, strict: bool) -> bool:
        """Check whether the actual filter is more or less specific than the cached filter.

        Args:
            strict: affects the result only when the two filters Actual and Cached are equal.
                If True, the filter Actual isn't considered a subfilter of Cached.
                If False, the filter Actual is considered a subfilter of Cached.

        Returns:
            True if the actual filter is more specific than the cached filter.
            False if the actual filter is less specific than the cached filter.
        """
        if not self._analysis_configs.cached:
            return False
        actual_filter = self._analysis_configs.actual.simulations_filter
        cached_filter = self._analysis_configs.cached.simulations_filter
        return is_subfilter(actual_filter, cached_filter, strict=strict)

    @_raise_if(locked=False)
    def repo_cache_needs_filter(self, name: str) -> bool:
        """Return True if the cached repo needs to be filtered.

        This happens when the cache is used, but the actual filter
        is more specific than the cached filter.
        """
        return self.is_repo_cached(name) and self._is_subfilter(strict=True)

    @_raise_if(locked=False)
    def features_cache_needs_filter(self, features_config: FeaturesConfig) -> bool:
        """Return True if the cached features need to be filtered.

        This happens when the cache is used, but the actual filter
        is more specific than the cached filter.
        """
        cached_checksums = self.get_cached_features_checksums(features_config)
        return len(cached_checksums) > 0 and self._is_subfilter(strict=True)



import pickle
import re

import pandas as pd
import pytest
from numpy.testing import assert_array_equal

from blueetl import repository as test_module
from blueetl.extract.neuron_classes import NeuronClasses
from blueetl.extract.neurons import Neurons
from blueetl.extract.simulations import Simulations
from blueetl.extract.spikes import Spikes
from blueetl.extract.windows import Windows
from tests.unit.utils import TEST_DATA_PATH, assert_frame_equal


def test_repository_init(repo):
    assert isinstance(repo, test_module.Repository)
    assert repo.extraction_config is not None
    assert repo.simulations_config is not None
    assert repo.cache_manager is not None
    assert repo.simulations_filter is not None
    assert repo.resolver is not None


def test_repository_names(repo):
    assert repo.names == ["simulations", "neurons", "neuron_classes", "windows", "report"]


def test_repository_extract(repo, capsys):
    with pytest.raises(RuntimeError, match="Not all the dataframes have been extracted"):
        repo.check_extractions()
    assert repo.is_extracted() is False
    repo.extract()
    assert repo.is_extracted() is True
    repo.check_extractions()
    repo.show()

    captured = capsys.readouterr()
    pattern = (
        "Extraction: simulations\n.*"
        "Extraction: neurons\n.*"
        "Extraction: neuron_classes\n.*"
        "Extraction: windows\n.*"
        "Extraction: report\n.*"
    )
    assert re.search(pattern, captured.out, flags=re.MULTILINE | re.DOTALL)


def test_repository_extract_simulations(repo):
    result = repo.simulations
    assert repo.simulation_ids == [0]
    assert isinstance(result, Simulations)
    expected_columns = [
        "seed",
        "simulation_path",
        "simulation_id",
        "circuit_id",
        "simulation",
        "circuit",
    ]
    assert_array_equal(result.df.columns, expected_columns)
    assert len(result.df) == 1

    missing_simulations = repo.missing_simulations()
    missing_simulation_path = str(
        TEST_DATA_PATH / "simulation_campaign" / "1" / "simulation_config.json"
    )
    expected_df = pd.DataFrame(
        data=[{"seed": 174404, "simulation_path": missing_simulation_path}],
        index=[1],
    )
    assert_frame_equal(missing_simulations, expected_df)


def test_repository_extract_neurons(repo):
    result = repo.neurons
    assert isinstance(result, Neurons)
    expected_columns = ["circuit_id", "neuron_class", "gid", "neuron_class_index"]
    assert_array_equal(result.df.columns, expected_columns)
    assert len(result.df) == 6


def test_repository_extract_neuron_classes(repo):
    result = repo.neuron_classes
    assert isinstance(result, NeuronClasses)
    expected_columns = [
        "circuit_id",
        "neuron_class",
        "count",
        "limit",
        "population",
        "node_set",
        "gids",
        "query",
    ]
    assert_array_equal(result.df.columns, expected_columns)
    assert len(result.df) == 3


def test_repository_extract_windows(repo):
    result = repo.windows
    assert isinstance(result, Windows)
    expected_columns = [
        "simulation_id",
        "circuit_id",
        "window",
        "trial",
        "offset",
        "t_start",
        "t_stop",
        "t_step",
        "duration",
        "window_type",
    ]
    assert_array_equal(result.df.columns, expected_columns)
    assert len(result.df) == 3


def test_repository_extract_report(repo):
    result = repo.report
    assert isinstance(result, Spikes)
    assert repo.spikes is result
    expected_columns = [
        "time",
        "gid",
        "window",
        "trial",
        "simulation_id",
        "circuit_id",
        "neuron_class",
    ]
    assert_array_equal(result.df.columns, expected_columns)
    assert len(result.df) == 16


def test_repository_pickle_roundtrip(repo):
    dumped = pickle.dumps(repo)
    loaded = pickle.loads(dumped)

    assert isinstance(loaded, test_module.Repository)
    assert loaded.extraction_config == repo.extraction_config
    assert loaded.simulations_config == repo.simulations_config
    assert loaded.simulations_filter == repo.simulations_filter
    assert loaded.names == repo.names


def test_repository_apply_filter(repo):
    filtered = repo.apply_filter({})
    assert isinstance(filtered, test_module.FilteredRepository)



"""Base data store."""

import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

import pandas as pd

from blueetl.types import StrOrPath
from blueetl.utils import checksum, resolve_path

L = logging.getLogger(__name__)


class BaseStore(ABC):
    """Abstract class defining a generic file data store.

    It's responsible for reading and writing Pandas DataFrames in a specific serialization format.
    """

    def __init__(self, basedir: StrOrPath) -> None:
        """Initialize the object.

        Args:
            basedir: base directory where the files should be stored.
        """
        self._basedir = resolve_path(basedir)
        L.debug("Using class %s with basedir %s", self.__class__.__name__, self.basedir)

    @property
    def basedir(self) -> Path:
        """Return the directory containing the files."""
        return self._basedir

    @property
    @abstractmethod
    def extension(self) -> str:
        """Return the file extension to be used with this specific data store."""

    @abstractmethod
    def dump(self, df: pd.DataFrame, name: str) -> None:
        """Save a dataframe to file, using the given name and the class extension."""

    @abstractmethod
    def load(self, name: str) -> Optional[pd.DataFrame]:
        """Load a dataframe from file, using the given name and the class extension."""

    def delete(self, name: str) -> None:
        """Delete the file with the given name and the class extension, if it exists."""
        self.path(name).unlink(missing_ok=True)

    def path(self, name: str) -> Path:
        """Return the full path of the file with the given name and the class extension.

        If name is empty, then return the base directory.
        This can be useful when working with partitioned DataFrames and
        directories containing multiple files.
        """
        return self.basedir / f"{name}.{self.extension}" if name else self.basedir

    def checksum(self, name: str) -> Optional[str]:
        """Return a checksum of the file, or None if it doesn't exist."""
        path = self.path(name)
        if path.is_file():
            return checksum(path)
        return None



import pandas as pd
import pytest

from blueetl import cache as test_module
from blueetl.campaign.config import SimulationCampaign
from tests.unit.utils import assert_frame_equal


@pytest.fixture
def cache_manager(global_config):
    simulations_config = SimulationCampaign.load(global_config.simulation_campaign)
    analysis_config = global_config.analysis["spikes"]
    cache_config = global_config.cache
    instance = test_module.CacheManager(
        cache_config=cache_config,
        analysis_config=analysis_config,
        simulations_config=simulations_config,
    )
    yield instance
    instance.close()


def test_lock_manager_exclusive(tmp_path):
    mode_ex = test_module.LockManager.LOCK_EX
    lock_manager_1 = test_module.LockManager(tmp_path)
    lock_manager_2 = test_module.LockManager(tmp_path)
    lock_manager_1.lock(mode_ex)
    # locking again shouldn't do anything
    lock_manager_1.lock(mode_ex)
    with pytest.raises(test_module.CacheError, match="Another process is locking"):
        lock_manager_2.lock(mode_ex)
    lock_manager_1.unlock()
    # unlocking again shouldn't do anything
    lock_manager_1.unlock()
    # now lock_manager_2 can lock and unlock
    lock_manager_2.lock(mode_ex)
    lock_manager_2.unlock()


def test_lock_manager_shared(tmp_path):
    mode_ex = test_module.LockManager.LOCK_EX
    mode_sh = test_module.LockManager.LOCK_SH
    lock_manager_1 = test_module.LockManager(tmp_path)
    lock_manager_2 = test_module.LockManager(tmp_path)
    lock_manager_3 = test_module.LockManager(tmp_path)
    # acquire a shared lock
    lock_manager_1.lock(mode_sh)
    # the lock can be shared
    lock_manager_2.lock(mode_sh)
    # but not acquired exclusively
    with pytest.raises(test_module.CacheError, match="Another process is locking"):
        lock_manager_3.lock(mode_ex)
    # not even from an already acquired lock
    with pytest.raises(test_module.CacheError, match="Another process is locking"):
        lock_manager_1.lock(mode_ex)
    # unless the other shared locks are released
    lock_manager_2.unlock()
    lock_manager_1.lock(mode_ex)
    lock_manager_1.unlock()


def test_cache_manager_init_and_close(global_config):
    simulations_config = SimulationCampaign.load(global_config.simulation_campaign)
    analysis_config = global_config.analysis["spikes"]
    cache_config = global_config.cache

    instance = test_module.CacheManager(
        cache_config=cache_config,
        analysis_config=analysis_config,
        simulations_config=simulations_config,
    )
    assert instance.locked is True
    assert instance.readonly is False
    instance.close()
    assert instance.locked is False
    assert instance.readonly is False


def test_cache_manager_to_readonly(global_config):
    simulations_config = SimulationCampaign.load(global_config.simulation_campaign)
    analysis_config = global_config.analysis["spikes"]
    cache_config = global_config.cache

    instance = test_module.CacheManager(
        cache_config=cache_config,
        analysis_config=analysis_config,
        simulations_config=simulations_config,
    )
    new_instance = instance.to_readonly()

    assert isinstance(instance, test_module.CacheManager)
    assert isinstance(new_instance, test_module.CacheManager)
    assert new_instance is not instance
    assert instance.readonly is False
    assert new_instance.readonly is True

    instance.close()
    with pytest.raises(
        test_module.CacheError,
        match=(
            "Method CacheManager.to_readonly cannot be called "
            "when the attributes are: {'locked': False}"
        ),
    ):
        instance.to_readonly()


def test_cache_manager_concurrency_is_not_allowed_when_locked(global_config):
    simulations_config = SimulationCampaign.load(global_config.simulation_campaign)
    analysis_config = global_config.analysis["spikes"]
    cache_config = global_config.cache

    instance = test_module.CacheManager(
        cache_config=cache_config,
        analysis_config=analysis_config,
        simulations_config=simulations_config,
    )
    # verify that a new instance cannot be created when the old instance is keeping the lock
    with pytest.raises(test_module.CacheError, match="Another process is locking"):
        test_module.CacheManager(
            cache_config=cache_config,
            analysis_config=analysis_config,
            simulations_config=simulations_config,
        )
    # verify that a new instance can be created after closing the old instance
    instance.close()
    instance = test_module.CacheManager(
        cache_config=cache_config,
        analysis_config=analysis_config,
        simulations_config=simulations_config,
    )
    instance.close()


def test_cache_manager_concurrency_is_allowed_when_readonly(global_config):
    simulations_config = SimulationCampaign.load(global_config.simulation_campaign)
    analysis_config = global_config.analysis["spikes"]
    cache_config = global_config.cache.model_copy(update={"readonly": False})
    cache_config_readonly = global_config.cache.model_copy(update={"readonly": True})

    # init the cache that will be used later
    instance = test_module.CacheManager(
        cache_config=cache_config,
        analysis_config=analysis_config,
        simulations_config=simulations_config,
    )
    instance.close()

    # use the same cache in multiple cache managers
    instances = [
        test_module.CacheManager(
            cache_config=cache_config_readonly,
            analysis_config=analysis_config,
            simulations_config=simulations_config,
        )
        for _ in range(3)
    ]
    for instance in instances:
        instance.close()


def test_cache_manager_clear_cache(global_config, tmp_path):
    simulations_config = SimulationCampaign.load(global_config.simulation_campaign)
    analysis_config = global_config.analysis["spikes"]
    cache_config = global_config.cache.model_copy(update={"clear": False})
    cache_config_clear = global_config.cache.model_copy(update={"clear": True})

    output = cache_config.path
    sentinel = output / "sentinel"

    assert output.exists() is False
    instance = test_module.CacheManager(
        cache_config=cache_config_clear,
        analysis_config=analysis_config,
        simulations_config=simulations_config,
    )
    instance.close()
    assert output.exists() is True
    assert sentinel.exists() is False
    sentinel.touch()

    # reuse the cache
    instance = test_module.CacheManager(
        cache_config=cache_config,
        analysis_config=analysis_config,
        simulations_config=simulations_config,
    )
    instance.close()
    assert output.exists() is True
    assert sentinel.exists() is True

    # delete the cache
    instance = test_module.CacheManager(
        cache_config=cache_config_clear,
        analysis_config=analysis_config,
        simulations_config=simulations_config,
    )
    instance.close()
    assert output.exists() is True
    assert sentinel.exists() is False


def test_cache_manager_dump_load_repo(cache_manager):
    assert cache_manager.is_repo_cached("simulations") is False
    assert cache_manager.load_repo("simulations") is None

    simulations_df = pd.DataFrame([{"simulation_id": 0, "circuit_id": 0}])
    cache_manager.dump_repo(simulations_df, name="simulations")

    assert cache_manager.is_repo_cached("simulations") is True

    loaded_simulations_df = cache_manager.load_repo("simulations")

    assert_frame_equal(loaded_simulations_df, simulations_df)


def test_cache_manager_dump_load_features(global_config, cache_manager):
    analysis_config = global_config.analysis["spikes"]
    features_config = analysis_config.features[0]

    assert cache_manager.load_features(features_config) is None

    features_dict = {
        "by_gid": pd.DataFrame([{"simulation_id": 0, "circuit_id": 0, "gid": 0}]),
    }
    cache_manager.dump_features(features_dict, features_config=features_config)

    loaded_features_dict = cache_manager.load_features(features_config)

    assert set(features_dict) == set(loaded_features_dict)
    assert_frame_equal(loaded_features_dict["by_gid"], features_dict["by_gid"])



"""Validation utilities."""

import importlib.resources
import logging

import jsonschema
import yaml

L = logging.getLogger(__name__)


class ValidationError(Exception):
    """Validation error."""


def read_schema(schema_name: str) -> dict:
    """Load a schema and return the result as a dictionary."""
    traversable = importlib.resources.files(__package__) / "schemas" / f"{schema_name}.yaml"
    with traversable.open(encoding="utf-8") as stream:
        return yaml.safe_load(stream)


def validate_config(config: dict, schema: dict) -> None:
    """Raise an exception if the configuration is not valid.

    Args:
        config: configuration to be validated.
        schema: json schema.

    Raises:
        ValidationError if the validation failed.
    """
    if config is None:
        L.error("The configuration cannot be empty.")
        raise ValidationError("Empty configuration")
    cls = jsonschema.validators.validator_for(schema)
    cls.check_schema(schema)
    validator = cls(schema)
    errors = list(validator.iter_errors(config))
    if errors:
        # Log an error message for each error.
        msg = []
        for n, e in enumerate(errors, 1):
            path = ".".join(str(elem) for elem in ["root"] + list(e.absolute_path))
            msg.append(f"{n}: Failed validating {path}: {e.message}")
        msg = "\n".join(msg)
        L.error("Invalid configuration:\n%s", msg)
        raise ValidationError("Invalid configuration")



from pathlib import Path

import pytest

from blueetl.adapters import node_sets as test_module
from tests.unit.utils import TEST_NODE_SETS_FILE, TEST_NODE_SETS_FILE_EXTRA, assert_isinstance


def test_node_sets_adapter():
    path = TEST_NODE_SETS_FILE
    expected_class_name = "bluepysnap.node_sets.NodeSets"

    obj = test_module.NodeSetsAdapter.from_file(path)

    assert_isinstance(obj.instance, expected_class_name)
    assert obj.exists() is True


def test_node_sets_adapter_update():
    path = TEST_NODE_SETS_FILE
    path_extra = TEST_NODE_SETS_FILE_EXTRA

    obj = test_module.NodeSetsAdapter.from_file(path)
    obj_extra = test_module.NodeSetsAdapter.from_file(path_extra)

    assert "Layer2" in obj.instance
    assert "ExtraLayer2" not in obj.instance
    assert "ExtraLayer2" in obj_extra.instance

    obj.update(obj_extra)
    assert "ExtraLayer2" in obj.instance


def test_node_sets_adapter_ior():
    path = TEST_NODE_SETS_FILE
    path_extra = TEST_NODE_SETS_FILE_EXTRA

    obj = test_module.NodeSetsAdapter.from_file(path)
    obj_extra = test_module.NodeSetsAdapter.from_file(path_extra)

    assert "Layer2" in obj.instance
    assert "ExtraLayer2" not in obj.instance
    assert "ExtraLayer2" in obj_extra.instance

    obj |= obj_extra
    assert "ExtraLayer2" in obj.instance


def test_node_sets_adapter_with_none():
    expected_class_name = "bluepysnap.node_sets.NodeSets"
    obj = test_module.NodeSetsAdapter.from_file(None)

    assert_isinstance(obj.instance, expected_class_name)
    assert obj.exists() is True
    assert obj.instance.content == {}


def test_node_sets_adapter_with_nonexistent_path():
    path = Path("path/to/non/existent/file.json")

    with pytest.raises(FileNotFoundError, match="No such file or directory"):
        test_module.NodeSetsAdapter.from_file(path)



"""Comparison features from BlueCV."""

import logging

from elephant.spike_train_dissimilarity import van_rossum_distance, victor_purpura_distance
from quantities import ms

L = logging.getLogger(__name__)


def get_VR(spiketrain1, spiketrain2, tau=10):
    """Get Van rossum distance between two spiketrains."""
    result = van_rossum_distance([spiketrain1, spiketrain2], tau * ms)[0, 1]
    # keep the real part because sometimes the result is complex?
    # example: SpikeTrain([4027.375, 4051.9, 4114.475], units="ms", t_stop=5000.0)
    return result.real


def get_VP(spiketrain1, spiketrain2, q=0.1):
    """Get Victor purpura distance between two spiketrains."""
    return victor_purpura_distance([spiketrain1, spiketrain2], q * (1 / ms))[0, 1]



"""Interfaces for Circuit."""

from abc import ABC, abstractmethod
from collections.abc import Mapping
from pathlib import Path
from typing import Generic, Optional, TypeVar

import numpy as np
import pandas as pd

from blueetl.adapters.interfaces.node_sets import NodeSetsInterface
from blueetl.utils import CachedPropertyMixIn

CircuitT = TypeVar("CircuitT")


class NodePopulationInterface(ABC):
    """NodePopulation Interface."""

    @abstractmethod
    def get(self, group, properties) -> pd.DataFrame:
        """Return a DataFrame of nodes."""

    @abstractmethod
    def ids(self, group) -> np.ndarray:
        """Return an array of node ids."""


class CircuitInterface(Generic[CircuitT], CachedPropertyMixIn, ABC):
    """Circuit Interface."""

    def __init__(self, circuit: CircuitT) -> None:
        """Init the circuit interface with the specified circuit."""
        self._circuit = circuit

    @property
    def instance(self) -> CircuitT:
        """Return the wrapped instance."""
        return self._circuit

    @abstractmethod
    def checksum(self) -> str:
        """Return a checksum of the relevant keys in the circuit configuration."""

    @property
    @abstractmethod
    def nodes(self) -> Mapping[Optional[str], NodePopulationInterface]:
        """Return the nodes as a dict: population -> nodes."""

    @property
    @abstractmethod
    def node_sets_file(self) -> Optional[Path]:
        """Returns the NodeSets file used by the circuit."""

    @property
    @abstractmethod
    def node_sets(self) -> NodeSetsInterface:
        """Returns the NodeSets used by the circuit."""



"""Bluepy circuit implementation."""

from collections.abc import Mapping
from pathlib import Path
from typing import Optional

from bluepy import Circuit

from blueetl.adapters.interfaces.circuit import CircuitInterface, NodePopulationInterface
from blueetl.adapters.interfaces.node_sets import NodeSetsInterface
from blueetl.utils import checksum_json


class CircuitImpl(CircuitInterface[Circuit]):
    """Bluepy circuit implementation."""

    def checksum(self) -> str:
        """Return a checksum of the relevant keys in the circuit configuration.

        Circuits are considered different in this context if any relevant key is different.
        """
        # the targets key isn't considered, since it may contain TargetFile defined in BlueConfig
        circuit_config_keys = [
            "cells",
            "morphologies",
            "morphology_type",
            "emodels",
            "mecombo_info",
            "connectome",
            "projections",
            "projections_metadata",
            "segment_index",
            "synapse_index",
            "atlas",
        ]
        return checksum_json({k: self._circuit.config.get(k) for k in circuit_config_keys})

    @property
    def nodes(self) -> Mapping[Optional[str], NodePopulationInterface]:
        """Return the nodes as a dict: population -> nodes.

        The population name in the returned dict is always None, because undefined in the config.
        """
        return {None: self._circuit.cells}

    @property
    def node_sets_file(self) -> Optional[Path]:
        """Returns the NodeSets file used by the circuit."""
        raise NotImplementedError

    @property
    def node_sets(self) -> NodeSetsInterface:
        """Returns the NodeSets used by the circuit."""
        raise NotImplementedError



"""Interfaces for NodeSets."""

from abc import ABC, abstractmethod
from typing import Generic, TypeVar

NodeSetsT = TypeVar("NodeSetsT")


class NodeSetsInterface(Generic[NodeSetsT], ABC):
    """NodeSets Interface."""

    def __init__(self, node_sets: NodeSetsT) -> None:
        """Init the NodeSets interface with the specified NodeSetsT."""
        self._node_sets = node_sets

    @property
    def instance(self) -> NodeSetsT:
        """Return the wrapped instance."""
        return self._node_sets

    @abstractmethod
    def update(self, other: "NodeSetsInterface") -> None:
        """Update the wrapped node sets."""



import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_series_equal


@pytest.fixture
def storable_df_with_unnamed_index():
    # Format: column -> (rows, dtype)
    data = {
        "a": ([0, 1, 2], int),
        "b": ([0, 1, 2.0], float),
        "c": ([np.nan, np.nan, np.nan], float),
        "d": ([0, None, np.nan], float),
        "e": ([None, None, None], object),
        "f": (["s0", "s1", "s2"], object),
        "g": (["s0", None, np.nan], object),
        "h": ([[1.1, 1.2], [5], None], object),
        "i": ([[1.1, 1.2], [], None], object),
        "j": ([[1.1, 1.2], [], [3, 4]], object),
        "k": ([[1.1, 1.2], [2], [3, 4]], object),
    }
    df = pd.DataFrame({k: np.array(v[0], dtype=v[1]) for k, v in data.items()})
    expected_dtypes = pd.Series({k: v[1] for k, v in data.items()})
    assert_series_equal(df.dtypes, expected_dtypes)
    return df


@pytest.fixture
def storable_df_with_named_index(storable_df_with_unnamed_index):
    df = storable_df_with_unnamed_index
    df.index.name = "i0"
    return df


@pytest.fixture
def storable_df_with_unnamed_multiindex(storable_df_with_unnamed_index):
    df = storable_df_with_unnamed_index
    df.index = pd.MultiIndex.from_tuples([(0, 0), (1, 0), (1, 1)])
    return df


@pytest.fixture
def storable_df_with_named_multiindex(storable_df_with_unnamed_index):
    df = storable_df_with_unnamed_index
    df.index = pd.MultiIndex.from_tuples([(0, 0), (1, 0), (1, 1)], names=["i0", "i1"])
    return df



import re
from itertools import chain

import pytest

from tests.functional.utils import TEST_DATA_PATH


def _get_marks(path, match_pattern, skip_pattern):
    """Return the marks for the specified path."""
    if skip_pattern and re.match(skip_pattern, path.name):
        return pytest.mark.skip(reason="File skipped because matching skip_pattern")
    if match_pattern and not re.match(match_pattern, path.name):
        return pytest.mark.skip(reason="File skipped because not matching match_pattern")
    return ()


def _get_analysis_configs(config_path, expected_path, match_pattern, skip_pattern):
    """Return a list of (analysis_config_path, expected_path) to be used as fixtures."""
    analysis_configs = [
        pytest.param(
            path,
            expected_path / path.stem.replace("analysis_config_", "analysis_"),
            marks=_get_marks(path, match_pattern, skip_pattern),
            id=str(path.relative_to(path.parents[2])),
        )
        for path in sorted(config_path.glob("analysis_config_*.yaml"))
    ]
    # ensure that there are config files to test
    assert len(analysis_configs) >= 1, f"No configuration files found in {config_path}"
    return analysis_configs


def pytest_addoption(parser):
    parser.addoption(
        "--analysis-config-dir",
        action="append",
        default=[],
        help="List of directories containing configurations to test",
    )
    parser.addoption(
        "--match-pattern",
        action="store",
        default="",
        help="Pattern matching the configuration files to be tested",
    )
    parser.addoption(
        "--skip-pattern",
        action="store",
        default="",
        help="Pattern matching the configuration files to be skipped",
    )
    parser.addoption(
        "--force-update",
        action="store_true",
        default=False,
        help="Update the expected files",
    )


def pytest_generate_tests(metafunc):
    # See https://docs.pytest.org/en/latest/example/parametrize.html#generating-parameters-combinations-depending-on-command-line
    if "analysis_config_path" in metafunc.fixturenames or "expected_path" in metafunc.fixturenames:
        parents = metafunc.config.getoption("analysis_config_dir") or ["sonata", "bbp"]
        match_pattern = metafunc.config.getoption("match_pattern")
        skip_pattern = metafunc.config.getoption("skip_pattern")
        analysis_configs = list(
            chain.from_iterable(
                _get_analysis_configs(
                    config_path=TEST_DATA_PATH / parent / "config",
                    expected_path=TEST_DATA_PATH / parent / "expected",
                    match_pattern=match_pattern,
                    skip_pattern=skip_pattern,
                )
                for parent in parents
            )
        )
        metafunc.parametrize("analysis_config_path, expected_path", analysis_configs)


def pytest_collection_modifyitems(config, items):
    # see https://docs.pytest.org/en/latest/example/simple.html#control-skipping-of-tests-according-to-command-line-option
    if config.getoption("force_update"):
        # do not skip any test
        return
    marker = pytest.mark.skip(reason="Not overwriting files, use --force-update if needed")
    for item in items:
        if "force_update" in item.keywords:
            item.add_marker(marker)


def pytest_configure(config):
    config.addinivalue_line("markers", "force_update: mark tests as updating the expected files")



"""Adapters for Circuit."""

from collections.abc import Mapping
from pathlib import Path
from typing import Optional

from blueetl.adapters.base import BaseAdapter
from blueetl.adapters.interfaces.circuit import CircuitInterface, NodePopulationInterface
from blueetl.adapters.node_sets import NodeSetsAdapter


class CircuitAdapter(BaseAdapter[CircuitInterface]):
    """Circuit Adapter."""

    @classmethod
    def from_file(cls, filepath: Optional[Path]) -> "CircuitAdapter":
        """Load and return a new object from file."""
        # pylint: disable=import-outside-toplevel
        if not filepath or not filepath.exists():
            return cls(None)
        CircuitImpl: type[CircuitInterface]
        if filepath.suffix == ".json":
            from blueetl.adapters.impl.bluepysnap.circuit import Circuit, CircuitImpl
        else:
            from blueetl.adapters.impl.bluepy.circuit import Circuit, CircuitImpl
        impl = CircuitImpl(Circuit(str(filepath)))
        return cls(impl)

    def checksum(self) -> str:
        """Return a checksum of the relevant keys in the circuit configuration."""
        return self._ensure_impl.checksum()

    @property
    def nodes(self) -> Mapping[Optional[str], NodePopulationInterface]:
        """Return the nodes as a dict: population -> nodes."""
        return self._ensure_impl.nodes

    @property
    def node_sets_file(self) -> Optional[Path]:
        """Returns the NodeSets file used by the circuit."""
        return self._ensure_impl.node_sets_file

    @property
    def node_sets(self) -> NodeSetsAdapter:
        """Returns the NodeSets file used by the circuit."""
        return NodeSetsAdapter(self._ensure_impl.node_sets)



"""Bluepysnap simulation implementation."""

from collections.abc import Mapping
from functools import cached_property
from pathlib import Path
from typing import Optional

from bluepysnap import Simulation

from blueetl.adapters.impl.bluepysnap.circuit import CircuitImpl
from blueetl.adapters.interfaces.circuit import CircuitInterface
from blueetl.adapters.interfaces.simulation import (
    PopulationReportInterface,
    PopulationSpikesReportInterface,
    SimulationInterface,
)


class SimulationImpl(SimulationInterface[Simulation]):
    """Bluepysnap simulation implementation."""

    def is_complete(self) -> bool:
        """Return True if the spikes can be loaded from the simulation, False otherwise.

        Used to ignore a simulation before the simulation campaign is complete.
        """
        config = self._simulation.spikes.config
        return Path(config.output_dir, config.spikes_file).exists()

    @cached_property
    def circuit(self) -> CircuitInterface:
        """Return the circuit used for the simulation."""
        return CircuitImpl(self._simulation.circuit)

    @property
    def spikes(self) -> Mapping[Optional[str], PopulationSpikesReportInterface]:
        """Return the spikes report as a dict: population -> report."""
        return self._simulation.spikes

    @property
    def reports(self) -> Mapping[str, Mapping[Optional[str], PopulationReportInterface]]:
        """Return the reports as a dict: name -> population -> report."""
        return self._simulation.reports



"""Calculate features for soma reports."""


def calculate_features_by_simulation_circuit(repo, key, df, params):
    """Calculate features for soma reports.

    The passed df must be grouped by simulation_id, circuit_id.
    """
    # pylint: disable=unused-argument
    assert key._fields == ("simulation_id", "circuit_id")
    groupby = ["neuron_class", "window"]
    by_neuron_class = df.groupby(groupby, observed=True)["value"].agg(["mean", "std"]).reset_index()
    return {
        "by_neuron_class": by_neuron_class,
    }



"""Convert a spikes file by generating synthetic circuit, simulation, and simulation campaign."""

import dataclasses
import logging
import math
from pathlib import Path

import h5py
import numpy as np
import pandas as pd

from blueetl.utils import dump_json, dump_yaml, import_optional_dependency, relpath, resolve_path

L = logging.getLogger(__name__)

TIMESTAMPS = "timestamps"
NODE_IDS = "node_ids"
DTYPES = {
    TIMESTAMPS: np.float64,
    NODE_IDS: np.uint64,
}


@dataclasses.dataclass
class DataStats:
    """Statistics on the imported spikes."""

    rows: int
    unique_ids: int
    min_id: int
    max_id: int
    min_timestamp: float
    max_timestamp: float


@dataclasses.dataclass
class OutputPaths:
    """Output paths."""

    base: Path

    @property
    def circuit_path(self) -> Path:
        """Return the circuit_path."""
        return self.base / "circuit"

    @property
    def simulation_path(self) -> Path:
        """Return the simulation_path."""
        return self.base / "simulation"

    @property
    def spikes_path(self) -> Path:
        """Return the spikes_path."""
        return self.simulation_path / "synthetic_spikes.h5"

    @property
    def nodes_path(self) -> Path:
        """Return the nodes_path."""
        return self.circuit_path / "synthetic_nodes.h5"

    @property
    def node_sets_path(self) -> Path:
        """Return the node_sets_path."""
        return self.circuit_path / "node_sets.json"

    @property
    def circuit_config_path(self) -> Path:
        """Return the circuit_config_path."""
        return self.circuit_path / "circuit_config.json"

    @property
    def simulation_config_path(self) -> Path:
        """Return the simulation_config_path."""
        return self.simulation_path / "simulation_config.json"

    @property
    def simulation_campaign_config_path(self) -> Path:
        """Return the simulation_campaign_config_path."""
        return self.base / "simulation_campaign_config.json"

    @property
    def analysis_config_path(self) -> Path:
        """Return the analysis_config_path."""
        return self.base / "analysis_config.yaml"

    def mkdirs(self):
        """Create the directories."""
        self.base.mkdir(exist_ok=True, parents=True)
        self.circuit_path.mkdir(exist_ok=True, parents=True)
        self.simulation_path.mkdir(exist_ok=True, parents=True)


def _load_csv(path: Path, sep: str = " ", **kwargs) -> pd.DataFrame:
    """Load and sort spikes data from csv file.

    Accepted column names: timestamps or times, node_ids or ids.
    """
    columns = [TIMESTAMPS, NODE_IDS]
    valid_columns = {*columns, "ids", "times"}
    df = pd.read_csv(path, sep=sep, usecols=lambda x: x in valid_columns, **kwargs)
    df = df.rename(columns={"ids": NODE_IDS, "times": TIMESTAMPS})
    if missing := set(columns).difference(df.columns):
        raise ValueError(f"Missing columns in the CSV file: {missing}")
    df = df[columns].sort_values(columns)
    L.info("Loaded file %s", path)
    return df


def _get_data_stats(df: pd.DataFrame) -> DataStats:
    """Calculate statistics on the imported spikes."""
    unique_ids = df[NODE_IDS].drop_duplicates()
    stats = DataStats(
        rows=len(df),
        unique_ids=len(unique_ids),
        min_id=unique_ids.min(),
        max_id=unique_ids.max(),
        min_timestamp=df[TIMESTAMPS].min(),
        max_timestamp=df[TIMESTAMPS].max(),
    )
    L.info(
        "CSV stats: rows=%s, unique_ids=%s, min_id=%s, max_id=%s, min_ts=%s, max_ts=%s",
        stats.rows,
        stats.unique_ids,
        stats.min_id,
        stats.max_id,
        stats.min_timestamp,
        stats.max_timestamp,
    )
    return stats


def _write_spikes(
    path: Path, node_population: str, timestamps: np.ndarray, node_ids: np.ndarray
) -> None:
    """Write a spikes file in SONATA format."""
    L.info("Writing %s", path)
    sorting_type = h5py.enum_dtype({"none": 0, "by_id": 1, "by_time": 2})
    with h5py.File(path, "w") as h5:
        root = h5.create_group("spikes")
        pop = root.create_group(node_population)
        pop.attrs.create("sorting", data=2, dtype=sorting_type)
        ts_dataset = pop.create_dataset(TIMESTAMPS, data=timestamps, dtype=DTYPES[TIMESTAMPS])
        ts_dataset.attrs.create("units", "ms")
        pop.create_dataset(NODE_IDS, data=node_ids, dtype=DTYPES[NODE_IDS])


def _write_circuit(path: Path, node_population: str, size: int) -> None:
    """Write a synthetic empty circuit in SONATA format."""
    L.info("Writing %s", path)
    voxcell = import_optional_dependency("voxcell")
    # CellCollection uses 1-based ids, since it predates SONATA
    nodes = pd.DataFrame(index=range(1, size + 1))
    nodes["_"] = np.zeros(size, dtype=np.int8)
    cc = voxcell.CellCollection.from_dataframe(nodes)
    cc.population_name = node_population
    cc.save(str(path))


def _write_node_sets(path: Path, node_population: str) -> None:
    """Write a partial node_sets_file."""
    L.info("Writing %s", path)
    node_sets = {
        "empty": {
            "population": node_population,
            "node_id": [],
        }
    }
    dump_json(path, node_sets)


def _write_circuit_config(
    path: Path, node_sets_path: Path, nodes_path: Path, node_population: str
) -> None:
    """Write a partial circuit config in SONATA format."""
    L.info("Writing %s", path)
    nodes_file = str(relpath(nodes_path, path.parent))
    node_sets_file = str(relpath(node_sets_path, path.parent))
    circuit_config = {
        "version": "2.4",
        "metadata": {"status": "partial"},
        "node_sets_file": node_sets_file,
        "networks": {
            "nodes": [
                {
                    "nodes_file": nodes_file,
                    "populations": {node_population: {}},
                }
            ],
            "edges": [],
        },
    }
    dump_json(path, circuit_config)


def _write_simulation_config(
    path: Path, circuit_config_path: Path, spikes_path: Path, tstop: float
) -> None:
    """Write a simulation config in SONATA format."""
    L.info("Writing %s", path)
    network = str(relpath(circuit_config_path, path.parent))
    output_dir = str(relpath(spikes_path.parent, path.parent))
    spikes_file = spikes_path.name
    simulation_config = {
        "version": "2.4",
        "network": network,
        "run": {
            "tstop": tstop,
            "dt": 1.0,
            "random_seed": 0,
        },
        "output": {
            "output_dir": output_dir,
            "spikes_file": spikes_file,
        },
        "metadata": {
            "note": "Synthetic simulation",
        },
    }
    dump_json(path, simulation_config)


def _write_simulation_campaign_config(
    path: Path, circuit_config_path: Path, simulation_config_path: Path
) -> None:
    """Write a simulation campaign config in BlueETL format, having only one simulation."""
    L.info("Writing %s", path)
    circuit_config = str(resolve_path(circuit_config_path))
    path_prefix = str(resolve_path(simulation_config_path.parents[1]))
    simulation_path = str(relpath(simulation_config_path, path_prefix))
    simulation_campaign_config = {
        "format": "blueetl",
        "version": 1,
        "name": "synthetic",
        "attrs": {
            "path_prefix": path_prefix,
            "circuit_config": circuit_config,
            "__coupled__": "coupled",
        },
        "data": [
            {"simulation_path": simulation_path},
        ],
    }
    dump_yaml(path, simulation_campaign_config)


def _write_analysis_config(path: Path, simulation_campaign_config_path: Path) -> None:
    """Write an analysis config in BlueETL format."""
    L.info("Writing %s", path)
    simulation_campaign = str(relpath(simulation_campaign_config_path, path.parent))
    analysis_config = {
        "version": 3,
        "simulation_campaign": simulation_campaign,
        "cache": {
            "path": "analysis",
        },
        "analysis": {
            "spikes": {
                "extraction": {
                    "report": {"type": "spikes"},
                    "neuron_classes": {
                        "all": {},
                    },
                    "limit": None,
                    "population": "synthetic",
                    "node_set": None,
                    "windows": {
                        "w1": {"bounds": [0.0, 1000.0]},
                        "w2": {"bounds": [1000.0, 2000.0]},
                    },
                },
                "features": [
                    {
                        "type": "multi",
                        "groupby": ["simulation_id", "circuit_id", "neuron_class", "window"],
                        "function": (
                            "blueetl.external.bnac.calculate_features.calculate_features_multi"
                        ),
                        "params": {"export_all_neurons": True},
                    }
                ],
            }
        },
    }
    dump_yaml(path, analysis_config)


def main(input_file: Path, output_dir: Path, node_population: str) -> None:
    """Read a CSV file containing the spikes, and write synthetic files to be used with BlueETL.

    Expected output files:

    ├── analysis_config.yaml
    ├── circuit
    │   ├── circuit_config.json
    │   ├── node_sets.json
    │   └── synthetic_nodes.h5
    ├── simulation
    │   ├── simulation_config.json
    │   └── synthetic_spikes.h5
    └── simulation_campaign_config.json

    """
    paths = OutputPaths(base=output_dir)
    paths.mkdirs()
    df = _load_csv(input_file)
    stats = _get_data_stats(df)

    _write_circuit(
        paths.nodes_path,
        node_population=node_population,
        size=stats.max_id + 1,
    )
    _write_node_sets(
        paths.node_sets_path,
        node_population=node_population,
    )
    _write_circuit_config(
        paths.circuit_config_path,
        nodes_path=paths.nodes_path,
        node_sets_path=paths.node_sets_path,
        node_population=node_population,
    )
    _write_spikes(
        paths.spikes_path,
        node_population=node_population,
        timestamps=df[TIMESTAMPS],
        node_ids=df[NODE_IDS],
    )
    _write_simulation_config(
        paths.simulation_config_path,
        circuit_config_path=paths.circuit_config_path,
        spikes_path=paths.spikes_path,
        tstop=float(math.ceil(stats.max_timestamp)),
    )
    _write_simulation_campaign_config(
        paths.simulation_campaign_config_path,
        circuit_config_path=paths.circuit_config_path,
        simulation_config_path=paths.simulation_config_path,
    )
    _write_analysis_config(
        paths.analysis_config_path,
        simulation_campaign_config_path=paths.simulation_campaign_config_path,
    )



"""Base extractor."""

import logging
from abc import ABC
from typing import Optional, TypeVar

import pandas as pd

from blueetl.utils import ensure_dtypes

L = logging.getLogger(__name__)
ExtractorT = TypeVar("ExtractorT", bound="BaseExtractor")


class BaseExtractor(ABC):
    """Base extractor class."""

    COLUMNS: list[str] = []
    _allow_missing_columns = False
    _allow_extra_columns = False
    _allow_empty_data = False

    def __init__(self, df: pd.DataFrame, cached: bool, filtered: bool) -> None:
        """Initialize the extractor.

        Args:
            df: Pandas DataFrame containing the extracted data.
            cached: True if the data have been extracted from the cache, False otherwise.
            filtered: True if the data have been filtered using a custom query, False otherwise.
        """
        self._cached = cached
        self._filtered = filtered
        self._validate(df)
        self._df = ensure_dtypes(df)

    @classmethod
    def _validate(cls, df: pd.DataFrame) -> None:
        """Validate the dataframe.

        It can be overridden if a custom validation is needed.
        """
        cls._validate_data(df)
        cls._validate_columns(df)

    @classmethod
    def _validate_data(cls, df: pd.DataFrame) -> None:
        """Validate the content of the dataframe."""
        if not cls._allow_empty_data and df.empty:
            raise RuntimeError(f"No data extracted to {cls.__name__}")

    @classmethod
    def _validate_columns(cls, df: pd.DataFrame) -> None:
        """Validate the names of the columns of the dataframe."""
        actual = set(df.columns)
        expected = set(cls.COLUMNS)
        if not cls._allow_missing_columns and expected - actual:
            raise ValueError(f"Expected columns not present: {expected - actual}")
        if not cls._allow_extra_columns and actual - expected:
            raise ValueError(f"Additional columns not allowed: {actual - expected}")

    @property
    def df(self) -> pd.DataFrame:
        """Return the internally wrapped dataframe."""
        return self._df

    @classmethod
    def from_pandas(
        cls: type[ExtractorT],
        df: pd.DataFrame,
        query: Optional[dict] = None,
        cached: bool = True,
    ) -> ExtractorT:
        """Return a new object from the given dataframe.

        If a query is specified, it's passed to ``etl.q`` and applied as a filter.

        It can be overridden together with ``to_pandas`` if some columns are not serializable.

        Args:
            df: dataframe to load.
            query: optional filter dictionary, passed to ``etl.q``.
            cached: True if the data is loaded from the cache, False otherwise.

        Returns:
            a new extractor instance.
        """
        original_len = len(df)
        if query:
            L.debug("Filtering dataframe %s by %s", cls.__name__, query)
            df = df.etl.q(query)
            if not any(df.index.names):
                # reset the index to remove any gap
                df = df.reset_index(drop=True)
        filtered = len(df) != original_len
        return cls(df, cached=cached, filtered=filtered)

    def to_pandas(self) -> pd.DataFrame:
        """Return a dataframe that can be serialized and stored to disk.

        It should be possible to call ``from_pandas`` with the returned dataframe to create
        an equivalent object.

        It can be overridden together with ``from_pandas`` if some columns are not serializable.

        Returns:
            serializable dataframe.
        """
        return self.df



from pathlib import Path
from unittest.mock import patch

from click.testing import CliRunner

from blueetl.apps import validate as test_module
from blueetl.validation import ValidationError


@patch(test_module.__name__ + ".validation.validate_config")
def test_validate_config_success(mock_validate_config, tmp_path):
    analysis_config_file = "config.yaml"
    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=tmp_path) as td:
        Path(analysis_config_file).write_text("---")
        result = runner.invoke(test_module.validate_config, [analysis_config_file])

    assert result.output.strip() == "Validation successful."
    assert result.exit_code == 0
    assert mock_validate_config.call_count == 1


@patch(test_module.__name__ + ".validation.validate_config")
def test_validate_config_failure(mock_validate_config, tmp_path):
    analysis_config_file = "config.yaml"
    runner = CliRunner()

    mock_validate_config.side_effect = ValidationError()
    with runner.isolated_filesystem(temp_dir=tmp_path) as td:
        Path(analysis_config_file).write_text("---")
        result = runner.invoke(test_module.validate_config, [analysis_config_file])

    assert result.output.strip() == "Validation failed."
    assert result.exit_code == 1
    assert mock_validate_config.call_count == 1



"""Common constants."""

import numpy as np

# columns names
GID = "gid"
TIME = "time"
SIMULATION_PATH = "simulation_path"
SIMULATION_ID = "simulation_id"
SIMULATION = "simulation"
CIRCUIT_ID = "circuit_id"
CIRCUIT = "circuit"
NEURON_CLASS = "neuron_class"
NEURON_CLASS_INDEX = "neuron_class_index"  # incremental gid index for each neuron class
WINDOW = "window"
TRIAL = "trial"
OFFSET = "offset"
T_START = "t_start"
T_STOP = "t_stop"
T_STEP = "t_step"
DURATION = "duration"
WINDOW_TYPE = "window_type"
COUNT = "count"
TIMES = "times"
BIN = "bin"
VALUE = "value"
SECTION = "section"
LIMIT = "limit"
POPULATION = "population"
NODE_SET = "node_set"
GIDS = "gids"
QUERY = "query"

DTYPES = {
    GID: np.int64,
    TIME: np.float64,
    SIMULATION_ID: np.int16,
    CIRCUIT_ID: np.int16,
    NEURON_CLASS: "category",
    WINDOW: "category",
    TRIAL: np.int16,
    OFFSET: np.float64,
    T_START: np.float64,
    T_STOP: np.float64,
    T_STEP: np.float64,
    DURATION: np.float64,
}
CHECKSUM_SEP = "#"
LEVEL_SEP = "."

CONFIG_VERSION = 4
MIN_SUPPORTED_CONFIG_VERSION = 3
